# coding: utf8
from __future__ import unicode_literals, print_function, division

from six import PY2
from nose.tools import assert_almost_equal, assert_equal
from clldutils.testing import capture

from pyglottolog.tests.util import WithApi


class Tests(WithApi):
    def test_roman(self):
        from pyglottolog.monsterlib.roman import introman, romanint

        self.assertEqual(introman(5), 'v')
        self.assertEqual(introman(8), 'viii')
        for i in range(1, 2000):
            self.assertEqual(i, romanint(introman(i)))

    def test_Collection(self):
        from pyglottolog.monsterlib._bibfiles_db import Database

        if not PY2:  # pragma: no cover
            return

        db = self.tmp_path('test.sqlite')
        with capture(self.api.bibfiles.to_sqlite, db) as out:
            self.assertIn('ENTRYTYPE', out)
        with capture(self.api.bibfiles.to_sqlite, db) as out:
            pass
        with capture(self.api.bibfiles.to_sqlite, db, rebuild=True) as out:
            pass

        db = Database(db, self.api.bibfiles)
        with capture(db.recompute, reload_priorities=self.api.bibfiles) as out:
            self.assertEqual(len(out.splitlines()), 34)
        with capture(db.is_uptodate, self.api.bibfiles[1:], verbose=True) as out:
            self.assertEqual(len(out.splitlines()), 3)
        db.to_bibfile(self.tmp_path('out.bib'))
        db.to_csvfile(self.tmp_path('out.csv'))
        db.to_replacements(self.tmp_path('out.json'))
        self.assertEqual(db.to_hhmapping(), {'s:Karang:Tati-Harzani': 41999})
        with capture(db.trickle) as out:
            self.assertIn('2 changed 1 added in a', out)
        key, (entrytype, fields) = db[('b.bib', 'arakawa97')]
        self.assertEqual(entrytype, 'article')
        self.assertEqual(fields['volume'], '16')
        with capture(db.stats) as out:
            pass

        for attr in ['splits', 'merges', 'identified', 'combined']:
            with capture(getattr(db, 'show_' + attr)) as out:
                pass

    def test_markconcservative(self):
        from pyglottolog.monsterlib._libmonster import markconservative

        res = markconservative(
            {1: ('article', {'title': 'Grammar'})},
            self.api.hhtypes.triggers,
            {1: ('article', {'title': 'Grammar'})},
            self.api.hhtypes,
            self.tmp_path('marks.txt'),
            verbose=False)
        self.assertEqual(res[1][1]['hhtype'].split()[0], 'grammar')

        # If a higher hhtype is computed, this cancels out previous computations.
        res = markconservative(
            {1: ('article', {'title': 'grammar', 'lgcode': 'abc'})},
            self.api.hhtypes.triggers,
            {1: ('article', {'title': 'other', 'hhtype': 'other', 'lgcode': 'abc'})},
            self.api.hhtypes,
            self.tmp_path('marks.txt'),
            verbose=False)
        self.assertNotIn('hhtype', res[1][1])

    def test_markall(self):
        from pyglottolog.monsterlib._libmonster import markall

        bib = {
            1: ('article', {'title': "other grammar of lang"}),
            2: ('article', {'title': "grammar of lang and dial"}),
            3: ('article', {'title': "other"}),
            4: ('article', {'title': "grammar and phonologie and morphologie"})
        }
        hht = self.api.hhtypes
        markall(bib, hht.triggers, verbose=False, rank=lambda l: hht[l].rank)
        self.assertIn('grammar', bib[1][1]['hhtype'])
        self.assertIn('morphologie and phonologie;grammar', bib[4][1]['hhtype'])

        markall(bib, self.api.triggers['lgcode'], verbose=False)
        self.assertIn('language', bib[1][1]['lgcode'])

    def test_add_inlg_e(self):
        from pyglottolog.monsterlib._libmonster import add_inlg_e, INLG

        res = add_inlg_e(
            {1: ('article', {'title': 'Grammar of language'})},
            self.api.triggers[INLG],
            verbose=False)
        assert_equal(res[1][1][INLG], 'language [abc]')


def test_names():
    from pyglottolog.monsterlib._bibtex import names

    assert_equal(
        [n.last for n in names('Alfred Meier and Peter von Bohr')],
        ['Meier', 'Bohr'])


def test_undiacritic():
    from pyglottolog.monsterlib._bibtex_undiacritic import undiacritic

    if not PY2:
        return  # pragma: no cover

    for i, o in [
        ("\\cmd{äöüß}", "aouss"),
    ]:
        assert_equal(undiacritic(i), o)


def test_ulatex_decode():
    from pyglottolog.monsterlib._bibtex_escaping import ulatex_decode

    if not PY2:
        return  # pragma: no cover

    for i, o, r in [
        ("", "", ""),
        ("&#97;", "a", ""),
        ("a\tb", "a\tb", ""),
        ("\\%\\&\\#", "%&#", "\\%\\&\\#"),
        ("a\\^o\=\,b", "aôb̦̄", ""),
        ("Luise\\~no", "Luiseño", ""),
        ("\\textdoublevertline", "‖", ""),
        ("\\url{abcdefg}", "abcdefg", ""),
        ("\\textdoublegrave{o}", "\u020d", ""),
        ("\\textsubu{\\'{v}}a", "v\u032e\u0301a", ""),
        ("ng\\~{\\;u}", "ngữ", ""),
        ('\germ \\"Uber den Wolken', "[deu] Über den Wolken", ""),
        ('P. V\\u{a}n-T\\;u\\;o', 'P. Văn-Tươ', ""),
        ('\\textit{\\"{u}bertext}', 'übertext', ""),
    ]:
        assert_equal(ulatex_decode(i), o)
        if r:
            assert_equal(o.encode('ulatex+utf8', errors='keep'), r)


def test_distance():
    from pyglottolog.monsterlib._bibfiles_db import distance

    assert_almost_equal(distance({}, {}), 0)
    d1 = dict(author='An Author', year='1998', title='The Title', ENTRYTYPE='article')
    for d2, dist in [
        ({}, 1.0),
        (d1, 0),
        (dict(author='An Author'), 0),
        (dict(author='Another Author'), 0.2173),
        (dict(author='An Author', title='Another Title'), 0.13636),
    ]:
        assert_almost_equal(distance(d1, d2), dist, places=3)


def test_keyid():
    from pyglottolog.monsterlib._libmonster import keyid

    for fields, res in [
        ({}, '__missingcontrib__'),
        (dict(author='An Author'), 'author_no-titlend'),
        (dict(editor='An Author'), 'author_no-titlend'),
        (dict(author='An Author', title='A rather long title'), 'author_rather-longnd'),
        (dict(author='An Author', title='Title', year='2014'), 'author_title2014'),
        (dict(author='An Author', volume='IV'), 'author_no-titleivnd'),
        (dict(author='An Author', extra_hash='a'), 'author_no-titlenda'),
    ]:
        assert_equal(keyid(fields, {}), res)

    with capture(keyid, dict(author='An Author and '), {}) as out:
        assert 'Unparsed' in out


def test_pyear():
    from pyglottolog.monsterlib._libmonster import pyear

    for year, res in [
        ('', '[nd]'),
        ('1931', '1931'),
        ('1931-32', '1931-1932'),
    ]:
        assert_equal(pyear(year), res)


def test_pagecount():
    from pyglottolog.monsterlib._libmonster import pagecount

    for pages, res in [
        ('', ''),
        ('1', '1'),
        ('10-20', '11'),
        ('10-20,v-viii', '4+11'),
        ('20,viii', '8+20'),
        ('10-2', '3'),  # interpreted as 10-12
    ]:
        assert_equal(pagecount(pages), res)


def test_lgcode():
    from pyglottolog.monsterlib._libmonster import lgcode

    for lgcode_, codes in [
        ('', []),
        ('[abc]', ['abc']),
        ('abc,NOCODE_Abc', ['abc', 'NOCODE_Abc']),
    ]:
        assert_equal(lgcode((None, dict(lgcode=lgcode_))), codes)


def test_grp2fd():
    from pyglottolog.monsterlib._libmonster import grp2fd

    l = [(1, 2), (1, 3), (2, 4), (1, 5)]
    assert_equal(grp2fd(l), {1: {2: 1, 3: 1, 5: 1}, 2: {4: 1}})



# _compare_names.py - compare regex-based with pybtex name parsing
from __future__ import print_function

import _bibfiles
import _bibtex
import _libmonster


def names1(s):  # pragma: no cover
    return [
        (debr(n.get('firstname', '')), debr(n.get('lastname', '')), debr(n.get('jr', '')))
        for n in _libmonster.pauthor(s)]


def names2(s):  # pragma: no cover
    return [
        (debr(first), debr(' '.join(n for n in (prelast, last) if n)), debr(lineage))
        for prelast, last, first, lineage in _bibtex.names(s)]


def debr(s):  # pragma: no cover
    if s.startswith('{') and s.endswith('}'):
        return s[1:-1]
    return s


if __name__ == '__main__':  # pragma: no cover
    from collections import Counter

    counts = Counter()
    for b in _bibfiles.Collection():
        print(b.filename.center(80, '#'))
        for bibkey, (entrytype, fields) in b.iterentries():
            for role in ('author', 'editor'):
                names = fields.get(role, '')
                n1, n2 = names1(names), names2(names)
                if n1 != n2:
                    counts.update([b.filename])
                    print(b.filename, bibkey, role)
                    print(repr(names))
                    print(n1)
                    print(n2)
                    print()

    print('\n'.join('%d\t%s' % (n, f) for f, n in sorted(counts.most_common())))



# coding: utf8
from __future__ import unicode_literals
from unittest import TestCase

from clldutils.testing import WithTempDir
from clldutils import jsonlib

from pyglottolog.tests.util import WithApi
from pyglottolog.languoids import Languoid, EndangermentStatus
from pyglottolog.objects import Glottocodes, Glottocode, Level, Country, Macroarea


class TestGlottocodes(WithTempDir):
    def test_Glottocodes(self):
        gcjson = self.tmp_path('glottocodes.json')
        jsonlib.dump({}, gcjson)

        glottocodes = Glottocodes(gcjson)
        gc = glottocodes.new('a', dry_run=True)
        self.assertTrue(gc.startswith('aaaa'))
        self.assertNotIn(gc, glottocodes)
        gc = glottocodes.new('a')
        self.assertIn(gc, glottocodes)
        # make sure it's also written to file:
        self.assertIn(gc, Glottocodes(gcjson))


class Tests(TestCase):
    def test_es(self):
        self.assertEqual(
            EndangermentStatus.critical,
            EndangermentStatus.get('Critically endangered'))

    def test_pattern(self):
        pattern = Glottocode.pattern
        for valid in [
            'abcd1234',
            'a12d3456',
        ]:
            self.assertIsNotNone(pattern.match(valid))

        for invalid in [
            'abcd123',
            '12d3456',
            'aNOCODE',
            'NOCODE_abd',
            'nocode',
        ]:
            self.assertIsNone(pattern.match(invalid))

    def test_Country(self):
        self.assertEqual(Country.from_text('Germany').id, 'DE')
        self.assertIsNone(Country.from_name('abcdefg'))
        self.assertIsNone(Country.from_id('abcdefg'))

    def test_init(self):
        with self.assertRaises(ValueError):
            Glottocode('a2')


class TestLanguoid(WithApi):
    def test_Level(self):
        self.assertGreater(Level.dialect, Level.language)
        self.assertEqual(Level.language, self.api.languoid('abcd1235').level)
        with self.assertRaises(ValueError):
            Level.get('abcde')

    def test_factory(self):
        f = Languoid.from_dir(self.api.tree.joinpath('abcd1234'))
        self.assertEqual(f.category, 'Family')
        l = Languoid.from_dir(self.api.tree.joinpath(f.id, 'abcd1235'))
        self.assertEqual(l.name, 'language')
        self.assertIn('abcd1235', repr(l))
        self.assertIn('language', '%s' % l)
        self.assertEqual(l.level, Level.language)
        self.assertAlmostEqual(l.latitude, 0.5)
        self.assertAlmostEqual(l.longitude, 0.5)
        l.latitude, l.longitude = 1.0, 1.0
        self.assertAlmostEqual(l.latitude, 1.0)
        self.assertAlmostEqual(l.longitude, 1.0)
        self.assertEqual(l.iso_code, 'abc')
        l.iso_code = 'cde'
        self.assertEqual(l.iso, 'cde')
        self.assertEqual(l.hid, 'abc')
        l.hid = 'abo'
        self.assertEqual(l.hid, 'abo')
        self.assertEqual(l.id, 'abcd1235')

        self.assertEqual(len(l.macroareas), 2)
        l.macroareas = [Macroarea.africa]
        self.assertEqual(l.macroareas, [Macroarea.africa])

        l.countries = self.api.countries[:2]
        self.assertEqual(len(l.countries), 2)

        self.assertEqual(l.parent, f)
        self.assertEqual(f.children[0], l)
        self.assertEqual(l.children[0].family, f)
        l.write_info(self.tmp_path().as_posix())
        self.assertTrue(self.tmp_path('abcd1235').exists())
        self.assertIsInstance(
            self.api.languoid('abcd1235').iso_retirement.asdict(), dict)
        self.assertIsNone(l.classification_comment)
        l.endangerment = 'Critically endangered'
        self.assertEqual(l.endangerment, EndangermentStatus.critical)
        self.assertEqual(l.names, {})
        l.cfg['altnames'] = {'glottolog': 'xyz'}
        self.assertIn('glottolog', l.names)
        self.assertEqual(l.identifier, {})
        l.cfg['identifier'] = {'multitree': 'xyz'}
        self.assertIn('multitree', l.identifier)

    def test_isolate(self):
        l = Languoid.from_dir(self.api.tree.joinpath('isol1234'))
        self.assertTrue(l.isolate)
        self.assertIsNone(l.parent)
        self.assertIsNone(l.family)

    def test_attrs(self):
        l = Languoid.from_name_id_level(
            self.api.tree, 'name', 'abcd1235', Level.language, hid='NOCODE')
        l.name = 'other'
        self.assertEqual(l.name, 'other')
        with self.assertRaises(AttributeError):
            l.glottocode = 'x'
        with self.assertRaises(AttributeError):
            l.id = 'x'
        self.assertEqual(l.id, l.glottocode)
        self.assertEqual(l.hid, 'NOCODE')



# coding: utf8
from __future__ import unicode_literals, print_function, division

from six import PY2
import attr
from whoosh import index
from whoosh.fields import Schema, TEXT, KEYWORD, ID, NUMERIC
from whoosh.analysis import StemmingAnalyzer
from whoosh.qparser import QueryParser, GtLtPlugin
from whoosh.highlight import Formatter, get_text

from clldutils.path import rmtree, as_unicode
from clldutils.misc import slug


@attr.s
class Languoid(object):
    id = attr.ib()
    iso = attr.ib()
    name = attr.ib()
    level = attr.ib()
    fname = attr.ib()
    highlights = attr.ib()


@attr.s
class Document(object):
    id = attr.ib()
    provider = attr.ib()
    title = attr.ib()
    author = attr.ib()
    authoryear = attr.ib()
    year = attr.ib()
    doctype = attr.ib()
    lgcode = attr.ib()


def get_langs_index(api, recreate=False):
    index_dir = api.build_path('whoosh_langs')
    if index_dir.exists() and recreate:
        rmtree(index_dir)  # pragma: no cover
    if not index_dir.exists():
        index_dir.mkdir()
        schema = Schema(
            id=ID(stored=True),
            name=TEXT(stored=True),
            fname=ID(stored=True),
            iso=ID(stored=True),
            level=KEYWORD(scorable=True, stored=True),
            macroarea=KEYWORD(scorable=True),
            country=KEYWORD(scorable=True),
            latitude=NUMERIC(),
            longitude=NUMERIC(),
            ini=TEXT(analyzer=StemmingAnalyzer(), stored=True),
        )
        return index.create_in(index_dir.as_posix(), schema)
    return index.open_dir(index_dir.as_posix())


class BracketFormatter(Formatter):
    """Puts square brackets around the matched terms.
    """

    def format_token(self, text, token, replace=False):
        tokentext = get_text(text, token, replace)
        return "[[%s]]" % tokentext


def search_langs(repos, q, limit=1000, **kw):
    index_ = get_langs_index(repos)
    qp = QueryParser("ini", schema=index_.schema)
    qp.add_plugin(GtLtPlugin())
    q = '{0} {1}'.format(q, ' '.join('{0}:"{1}"'.format(k, v) for k, v in kw.items()))

    def highlight(res):
        hl = res.highlights('ini', top=1)
        if hl:
            for line in hl.split('\n'):
                if '[[' in line:
                    return line.strip()

    with index_.searcher() as searcher:
        results = searcher.search(qp.parse(q), limit=limit)
        results.formatter = BracketFormatter()
        return (
            len(results),
            [
                Languoid(
                    r['id'], r.get('iso'), r['name'], r['level'], r['fname'], highlight(r)
                ) for r in results])


def build_langs_index(api, log):
    writer = get_langs_index(api, recreate=True).writer()
    for lang in api.languoids():
        writer.add_document(
            id=lang.id,
            name=lang.name,
            fname=as_unicode(lang.fname),
            iso=lang.iso,
            level=lang.level.name.decode() if PY2 else lang.level.name,
            macroarea=' '.join('{0}'.format(ma) for ma in lang.macroareas),
            country=' '.join('{0}'.format(c) for c in lang.countries),
            latitude=lang.latitude,
            longitude=lang.longitude,
            ini=lang.cfg.write_string(),
        )
    writer.commit()


def get_index(api, recreate=False):
    index_dir = api.ftsindex
    if index_dir.exists() and recreate:
        rmtree(index_dir)  # pragma: no cover
    if not index_dir.exists():
        index_dir.mkdir()
        schema = Schema(
            id=ID(stored=True),
            provider=KEYWORD(stored=True),
            authoryear=TEXT(stored=True),
            title=TEXT(analyzer=StemmingAnalyzer(), stored=True),
            author=TEXT(stored=True),
            year=TEXT(stored=True),
            doctype=TEXT(stored=True),
            lgcode=TEXT(stored=True),
            body=TEXT(),
            tags=KEYWORD)
        return index.create_in(index_dir.as_posix(), schema)
    return index.open_dir(index_dir.as_posix())


def search(repos, q, limit=1000, **kw):
    index_ = get_index(repos)
    qp = QueryParser("body", schema=index_.schema)
    q = '{0} {1}'.format(q, ' '.join('{0}:"{1}"'.format(k, v) for k, v in kw.items()))

    with index_.searcher() as searcher:
        results = searcher.search(qp.parse(q), limit=limit)
        return len(results), [Document(**res) for res in results]


def build_index(api, log):
    writer = get_index(api, recreate=True).writer()
    for bibfile in api.bibfiles:
        log.info('indexing {0}'.format(bibfile))
        for e in bibfile.iterentries():
            author = e.fields.get('author', '')
            if author:
                author = slug(author.split()[0])
            writer.add_document(
                id='{0}:{1}'.format(bibfile.fname.stem, e.key),
                provider='%s' % bibfile.fname.stem,
                title=e.fields.get('title', e.fields.get('booktitle', '')),
                author=e.fields.get('author', e.fields.get('editor', '')),
                year=e.fields.get('year', ''),
                doctype=e.fields.get('hhtype', ''),
                lgcode=e.fields.get('lgcode', ''),
                body='%s' % e.fields,
                authoryear='{0}{1}'.format(author, e.fields.get('year', '')).lower())
    writer.commit()



# coding: utf8
from __future__ import unicode_literals, print_function, division

from mock import patch, Mock

from pyglottolog.tests.util import WithApi


class Tests(WithApi):
    def test_bibtex(self):
        from pyglottolog.iso import bibtex

        with patch(
                'pyglottolog.iso.requests',
                Mock(get=Mock(return_value=Mock(content=HTML)))):
            self.assertEqual(bibtex(self.api, Mock()), 2)


HTML = """\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<body>
<div id="main">
<h1>ISO 639-3 Index of All Past Change Requests</h1>
<table  width="100%" class="stripeMe">
<tr VALIGN="TOP">
   <th ALIGN="LEFT">CR Number</th>
   <th ALIGN="LEFT">Region</th>
   <th ALIGN="LEFT">Language Family</th>
   <th ALIGN="LEFT">Affected Identifier</th>
   <th ALIGN="LEFT">Language Name</th>
   <th ALIGN="LEFT">Change Type</th>
		<th ALIGN="LEFT">Outcome/<br>Effective date</th>
   <th ALIGN="LEFT" WIDTH="10%"></th>
</tr>
	<tr VALIGN="TOP">
		<td>2006-001</td>
		<td>Europe, Western</td>
		<td>Deaf sign language</td>
		<td>bvs</td>
		<td>
			Belgian Sign Language
		</td>
		<td>Split</td>
		<td>Adopted<br>2007-07-18 </td>
		<td><a HREF="chg_detail.asp?id=2006-001&lang=bvs">more ...</a></td>
	</tr>
	<tr VALIGN="TOP">
		<td>2006-001</td>
		<td>Europe, Western</td>
		<td>Deaf sign language</td>
		<td>sfb</td>
		<td>
			Langue des signes de Belgique Francophone
		</td>
		<td>Create</td>
		<td>Adopted<br>2007-07-18 </td>
		<td><a HREF="chg_detail.asp?id=2006-001&lang=sfb">more ...</a></td>
	</tr>
	<tr VALIGN="TOP">
		<td>2006-023</td>
		<td>Asia, South</td>
		<td>Kuki-Chin-Naga</td>
		<td>flm</td>
		<td>
			Falam Chin
		</td>
		<td>Split</td>
		<td>Adopted<br>2007-07-18 </td>
		<td><a HREF="chg_detail.asp?id=2006-023&lang=flm">more ...</a></td>
	</tr>
</table>
</div>
</body>
</html>
"""



# coding: utf8
from __future__ import unicode_literals, print_function, division
import re
import functools
from itertools import chain
from collections import Counter, OrderedDict
import unicodedata
import datetime

from six import string_types
import attr
from clldutils.misc import cached_property, UnicodeMixin
from clldutils.path import memorymapped
from clldutils.source import Source

from pyglottolog.util import Trigger
from pyglottolog.monsterlib import _bibtex
from pyglottolog.monsterlib._bibfiles_db import Database


class BibFiles(list):
    """Directory with an INI-file with settings for BibTeX files inside."""
    def __init__(self, api, ini):
        res = []
        for sec in ini.sections():
            if sec.endswith('.bib'):
                fname = api.references_path('bibtex', sec)
                if not fname.exists():  # pragma: no cover
                    raise ValueError('invalid bibtex file referenced in BIBFILES.ini')
                res.append(BibFile(fname=fname, **ini[sec]))

        super(BibFiles, self).__init__(res)
        self._map = {b.fname.name: b for b in self}

    def __getitem__(self, index_or_filename):
        """Retrieve a bibfile by index or filename."""
        if isinstance(index_or_filename, string_types):
            return self._map[index_or_filename]
        return super(BibFiles, self).__getitem__(index_or_filename)

    def to_sqlite(self, filename, rebuild=False):
        """Return a database with the bibfiles loaded."""
        return Database.from_bibfiles(self, filename, rebuild=rebuild)

    def roundtrip_all(self):
        """Load and save all bibfiles with the current settings."""
        return [b.roundtrip() for b in self]


@attr.s
class Entry(UnicodeMixin):
    key = attr.ib()
    type = attr.ib()
    fields = attr.ib()
    bib = attr.ib()

    # FIXME: add method to apply triggers!

    lgcode_regex = '[a-z0-9]{4}[0-9]{4}|[a-z]{3}|NOCODE_[A-Z][^\s\]]+'
    lgcode_in_brackets_pattern = re.compile("\[(" + lgcode_regex + ")\]")
    recomma = re.compile("[,/]\s?")
    lgcode_pattern = re.compile(lgcode_regex + "$")

    def __unicode__(self):
        res = "@%s{%s" % (self.type, self.key)
        for k, v in _bibtex.fieldorder.itersorted(self.fields):
            res += ',\n    %s = {%s}' % (k, v.strip() if hasattr(v, 'strip') else v)
        res += '\n}\n' if self.fields else ',\n}\n'
        return res

    def text(self):
        return Source(self.type, self.key, _check_id=False, **self.fields).text()

    @property
    def id(self):
        return '{0}:{1}'.format(self.bib.id, self.key)

    @classmethod
    def lgcodes(cls, string):
        if string is None:
            return []
        codes = cls.lgcode_in_brackets_pattern.findall(string)
        if not codes:
            # ... or as comma separated list of identifiers.
            parts = [p.strip() for p in cls.recomma.split(string)]
            codes = [p for p in parts if cls.lgcode_pattern.match(p)]
            if len(codes) != len(parts):
                codes = []
        return codes

    def iterlanguoids(self, langs_by_codes):
        if 'lgcode' in self.fields:
            for code in self.lgcodes(self.fields['lgcode']):
                if code in langs_by_codes:
                    yield langs_by_codes[code]


def file_if_exists(i, a, value):
    if value.exists() and not value.is_file():
        raise ValueError('invalid path')  # pragma: no cover


@attr.s
class BibFile(UnicodeMixin):
    fname = attr.ib(validator=file_if_exists)
    name = attr.ib(default=None)
    title = attr.ib(default=None)
    description = attr.ib(default=None)
    abbr = attr.ib(default=None)
    encoding = attr.ib(default='utf-8-sig')
    sortkey = attr.ib(
        default=None, convert=lambda s: None if s is None or s.lower() == 'none' else s)
    priority = attr.ib(default=0, convert=int)
    url = attr.ib(default=None)

    @property
    def id(self):
        return self.fname.stem

    def __getitem__(self, item):
        if item.startswith(self.id + ':'):
            item = item.split(':', 1)[1]
        text = None
        with memorymapped(self.fname) as string:
            m = re.search(b'@[A-Za-z]+\{' + re.escape(item.encode('utf8')), string)
            if m:
                next = string.find(b'\n@', m.end())
                if next >= 0:
                    text = string[m.start():next]
                else:
                    text = string[m.start():]
        if text:
            for k, (t, f) in _bibtex.iterentries_from_text(text, encoding=self.encoding):
                return Entry(k, t, f, self)
        raise KeyError(item)

    def visit(self, visitor=None):
        entries = OrderedDict()
        for entry in self.iterentries():
            if visitor is None or visitor(entry) is not True:
                entries[entry.key] = (entry.type, entry.fields)
        self.save(entries)

    @property
    def size(self):
        return self.fname.stat().st_size

    @property
    def mtime(self):
        return datetime.datetime.fromtimestamp(self.fname.stat().st_mtime)

    def iterentries(self):
        for k, (t, f) in _bibtex.iterentries(filename=self.fname, encoding=self.encoding):
            yield Entry(k, t, f, self)

    def keys(self):
        return ['{0}:{1}'.format(self.id, e.key) for e in self.iterentries()]

    @property
    def glottolog_ref_id_map(self):
        return {
            e.key: e.fields['glottolog_ref_id'] for e in self.iterentries()
            if 'glottolog_ref_id' in e.fields}

    def update(self, fname):
        entries = OrderedDict()
        ref_id_map = self.glottolog_ref_id_map
        for key, (type_, fields) in _bibtex.iterentries(fname, self.encoding):
            if key in ref_id_map and 'glottolog_ref_id' not in fields:
                fields['glottolog_ref_id'] = ref_id_map[key]
            entries[key] = (type_, fields)
        self.save(entries)

    def load(self):
        """Return entries as bibkey -> (entrytype, fields) dict."""
        return _bibtex.load(
            self.fname, preserve_order=self.sortkey is None, encoding=self.encoding)

    def save(self, entries):
        """Write bibkey -> (entrytype, fields) map to file."""
        _bibtex.save(
            entries, filename=self.fname, sortkey=self.sortkey, encoding=self.encoding)

    def __unicode__(self):
        return '<%s %s>' % (self.__class__.__name__, self.fname.name)

    def check(self, log):
        entries = self.load()  # bare BibTeX syntax
        invalid = _bibtex.check(filename=self.fname)  # names/macros etc.
        verdict = ('(%d invalid)' % invalid) if invalid else 'OK'
        method = log.warn if invalid else log.info
        method('%s %d %s' % (self, len(entries), verdict))
        return len(entries), verdict

    def roundtrip(self):
        print(self)
        self.save(self.load())

    def show_characters(self, include_plain=False):
        """Display character-frequencies (excluding printable ASCII)."""
        with self.fname.open(encoding=self.encoding) as fd:
            hist = Counter(fd.read())
        table = '\n'.join(
            '%d\t%-9r\t%s\t%s' % (n, c, c, unicodedata.name(c, ''))
            for c, n in hist.most_common()
            if include_plain or not 20 <= ord(c) <= 126)
        print(table)


@functools.total_ordering
class HHType(object):
    def __init__(self, s, p):
        self.name = s
        self.id = p.get(s, 'id')
        self.rank = p.getint(s, 'rank')
        self.abbv = p.get(s, 'abbv')
        self.bibabbv = p.get(s, 'bibabbv')
        self.triggers = [Trigger('hhtype', self.id, t)
                         for t in p.get(s, 'triggers').strip().splitlines() or []]

    def __repr__(self):
        return '<%s %s rank=%s>' % (self.__class__.__name__, self.id, self.rank)

    def __eq__(self, other):
        return self.rank == other.rank

    def __lt__(self, other):
        return self.rank < other.rank


class HHTypes(object):
    _rekillparen = re.compile(" \([^\)]*\)")
    _respcomsemic = re.compile("[;,]\s?")

    def __init__(self, ini):
        self._types = sorted([HHType(s, ini) for s in ini.sections()], reverse=True)
        self._type_by_id = {t.id: t for t in self._types}

    @classmethod
    def parse(cls, s):
        return cls._respcomsemic.split(cls._rekillparen.sub("", s))

    def __iter__(self):
        return iter(self._types)

    def __len__(self):
        return len(self._types)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._types[0]
        return self._type_by_id.get(item, self._type_by_id.get('unknown'))

    @cached_property()
    def triggers(self):
        return list(chain(*[t.triggers for t in self]))



# coding: utf8
from __future__ import unicode_literals, print_function, division
import os
from glob import glob
import io
from csv import writer


def read(fp):
    res = []
    for line in fp:
        cols = line.split(',')
        res.append(cols)
    return res


def walk(d):
    for fname in os.path.join(d, '*.csv'):
        data = read(fname)



# coding: utf8
from __future__ import unicode_literals, print_function, division

from clldutils.testing import capture

from pyglottolog.tests.util import WithApi


class Tests(WithApi):
    def test_paths(self):
        self.assert_(self.api.ftsindex)

    def test_languoid(self):
        self.assertEqual(self.api.languoid('abc').name, 'language')

    def test_languoids(self):
        from pyglottolog.languoids import Level

        self.assertEqual(len(list(self.api.languoids())), 4)
        self.assertEqual(len(list(self.api.languoids(maxlevel=Level.family))), 1)
        self.assertEqual(len(list(self.api.languoids(maxlevel=Level.language))), 3)
        self.assertEqual(len(self.api.languoids_by_code()), 7)
        self.assertIn('NOCODE_Family-name', self.api.languoids_by_code())

    def test_load_triggers(self):
        self.assertEqual(len(self.api.triggers), 2)

    def test_macroarea_map(self):
        self.assertEqual(self.api.macroarea_map['abc'], 'Eurasia')

    def test_bibfiles(self):
        c = self.api.bibfiles
        with capture(c.roundtrip_all) as out:
            self.assertIn('a.bib', out)
        with capture(c['b.bib'].show_characters) as out:
            self.assertIn('CJK UNIFIED IDEOGRAPH', out)
        abib = c[0]
        self.assertEqual(len(list(abib.iterentries())), 3)
        assert abib.size
        assert abib.mtime



# coding: utf8
from __future__ import unicode_literals, print_function, division

from unittest import TestCase


class Tests(TestCase):
    def test_unique(self):
        from pyglottolog.util import unique

        l = [1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 6]
        self.assertEqual(len(set(l)), len(list(unique(l))))

    def test_wrap(self):
        from pyglottolog.util import wrap

        text = """\
This is going to be a long line which must be split.
The next line is going to be interpreted as second paragraph upon first pass.
"""
        wrapped = wrap(text, width=20, line_as_paragraph=True)

        # Now a second pass should be idempotent:
        self.assertEqual(wrapped, wrap(wrapped, width=20))

    def test_group_first(self):
        from pyglottolog.util import group_first

        for key, items in group_first([(1, 2), (1, 3)]):
            self.assertEqual(key, 1)
            self.assertEqual(len(list(items)), 2)
            break

    def test_Trigger(self):
        from pyglottolog.util import Trigger

        t1 = Trigger('hhtype', 'grammar', 'phonologie AND NOT morphologie')
        t2 = Trigger('hhtype', 'phonology', 'phonologie')
        t3 = Trigger('hhtype', 'grammar', 'grammar')

        assert t1 != t3 and t1 == t1
        allkeys = range(5)
        keys_by_word = dict(grammar=[1, 2], phonologie=[2, 3], morphologie=[3, 4])
        self.assertEqual(t1(allkeys, keys_by_word), {2})
        self.assertEqual(t2(allkeys, keys_by_word), {2, 3})
        self.assertEqual(t3(allkeys, keys_by_word), {1, 2})
        self.assertIn('not morphologie and phonologie', Trigger.format('a', t1))

        for t in sorted([t1, t2, t3]):
            self.assertIn(t.type, Trigger.format(t.type, t))



# coding: utf8
from __future__ import unicode_literals

from mock import Mock
from clldutils.path import walk

from pyglottolog.lff import read_lff, rmtree, lff2tree, tree2lff
from pyglottolog.languoids import Level
from pyglottolog.tests.util import WithApi


class TestsWithFiles(WithApi):
    def test_rmtree(self):
        comps = []
        for i in range(100):
            comps.append('a')
            d = self.tmp_path(*comps)
            d.mkdir()
            f = d.joinpath('a.ini')
            with f.open('w', encoding='utf8') as fp:
                fp.write('a')
        assert self.tmp_path('a').exists()
        rmtree(self.tmp_path('a'))
        assert not self.tmp_path('a').exists()

    def _set_lff(self, content, name):
        with self.api.build_path(name).open('w', encoding='utf8') as fp:
            fp.write(content)
        return content

    def test_lff2tree(self):
        lfftext = self._set_lff("""# -*- coding: utf-8 -*-
Abkhaz-Adyge [abkh1242] aaa
    Ubykh [ubyk1235]uby
Abkhaz-Adyge [abkh1242] aaa; Abkhaz-Abaza [abkh1243]
    Abaza [abaz1241]abq
    Abkhazian [abkh1244]abk
Abkhaz-Adyge [abkh1242] aaa; Circassian [circ1239]
    Adyghe [adyg1241]ady
    Kabardian [kaba1278]kbd
""", 'lff.txt')

        dfftext = self._set_lff("""# -*- coding: utf-8 -*-
Abaza [abaz1241] abq
    Ashkaraua [ashk1247]
    Bezshagh [bezs1238]
    Tapanta [tapa1256]
Abkhazian [abkh1244] abk
    Abzhui [abzh1238]
    Bzyb [bzyb1238]
    Samurzakan [samu1242]
""", 'dff.txt')

        lff2tree(self.api)
        self.assertEqual(self.api.languoid('abkh1242').iso, 'aaa')
        self.assertEqual(self.api.languoid('ashk1247').level, Level.dialect)
        self.assertEqual(self.api.languoid('abaz1241').level, Level.language)
        self.assertEqual(self.api.languoid('abaz1241').hid, 'abq')

        self._set_lff(lfftext.replace('Abkhaz-Abaza', 'Abkhaz-Abazzza'), 'lff.txt')
        lff2tree(self.api)
        glottocodes = [d.name for d in walk(self.api.tree, mode='dirs')]
        self.assertEqual(len(glottocodes), len(set(glottocodes)))
        self.assertEqual(self.api.languoid('abkh1243').name, 'Abkhaz-Abazzza')

        lfftext = self._set_lff("""# -*- coding: utf-8 -*-
Abkhaz-Adyge [abkh1242]
    Ubykh [ubyk1235]
Abkhaz-Adyge [abkh1242]; Abkhaz-Abaza [abkh1243]; Abaza [abaz1241]
    Ashkaraua [ashk1247]xyz
    Abkhazian [abkh1244]
Abkhaz-Adyge [abkh1242]; Circassian [circ1239]
    Adyghe [adyg1241]ady
    Kabardian [kaba1278]
Abkhaz-Adyge [abkh1242]; Circassian [circ1239]; New Group []
    New name []NOCODE_New-name
    Another one []
""", 'lff.txt')

        self._set_lff("""# -*- coding: utf-8 -*-
Ashkaraua [ashk1247]xyz
    Bezshagh [bezs1238]
    Tapanta [tapa1256]
Abkhazian [abkh1244]
    Abzhui [abzh1238]
    Bzyb [bzyb1238]
    Samurzakan [samu1242]
Kabardian [kaba1278]
    Dia []aaa
""", 'dff.txt')

        lff2tree(self.api)
        self.assertEqual(self.api.languoid('abaz1241').level, Level.family)
        # Now we test two things:
        # - aaa has been removed as ISO code from abkh1242
        # - aaa has been attached as ISO code to a newly created language
        self.assertEqual(self.api.languoid('aaa').name, 'Dia')
        langs = list(self.api.languoids())
        self.assertIn('newg1234', self.api.glottocodes)
        self.assertEqual(len([l for l in langs if l.name == 'New Group']), 1)
        self.assertEqual(len([l for l in langs if l.hid == 'NOCODE_New-name']), 1)

        # Test ISO code removal:
        self._set_lff("""# -*- coding: utf-8 -*-
Kabardian [kaba1278]
    Dia []
""", 'dff.txt')
        lff2tree(self.api)
        self.assertIsNone(self.api.languoid('aaa'))

        tree2lff(self.api)

        #
        # Nodes must have unique names!
        #
        self._set_lff("""# -*- coding: utf-8 -*-
Ashkaraua [ashk1247]
    Ashkaraua [bezs1238]
""", 'dff.txt')
        with self.assertRaisesRegexp(ValueError, 'duplicate'):
            lff2tree(self.api)

        #
        # Nodes must have consistent names!
        #
        self._set_lff("""# -*- coding: utf-8 -*-
Ashkxxxaraua [ashk1247]
    Bezshagh [bezs1238]
""", 'dff.txt')
        with self.assertRaisesRegexp(ValueError, 'inconsistent'):
            lff2tree(self.api)

        #
        # Top-level nodes in dff must be languages:
        #
        self._set_lff("""# -*- coding: utf-8 -*-
Abaza [abaz1241]
    Bezshagh [bezs1238]
""", 'dff.txt')
        with self.assertRaises(ValueError):
            lff2tree(self.api)

        #
        # Top-level nodes in dff must be languages in lff:
        #
        self._set_lff("""# -*- coding: utf-8 -*-
None [xyzz1234]
    Dia []
""", 'dff.txt')
        with self.assertRaises(ValueError):
            lff2tree(self.api)

        #
        # Isolates must not have multiple ancestors:
        #
        self._set_lff("""# -*- coding: utf-8 -*-
None [xyzz1234]; Other [-isolate-]
    Dia []
""", 'dff.txt')
        with self.assertRaisesRegexp(ValueError, 'isolate'):
            lff2tree(self.api)

        #
        # Languages must appear after a classification line:
        #
        self._set_lff("""# -*- coding: utf-8 -*-
    Dia []
""", 'dff.txt')
        with self.assertRaisesRegexp(ValueError, 'classification'):
            lff2tree(self.api)

    def test_read_lff_error(self):
        _lff = """
Name [ac1234]; Name2 [abcd1235]
   Lang [abcd1236]abc
"""
        with self.assertRaises(ValueError):
            list(read_lff(self.api, Mock(), {}, Level.language, _lff.split('\n')))



# coding: utf-8
"""
Unfortunately out bibtex source files are not encoded consistently across files. I.e.
latex commands with the same name may be used with different definitions in different
files.

So far identified problems:
- \:n is used in sala.bib for N WITH DOT ABOVE, but is also defined in the textipa
  package, commands of which have been used in various other source files.


See also:
- textipa package reference
  http://www.ling.ohio-state.edu/~mbeckman/ling600.01/tipachart.pdf
- comprehensive latex symbol list
  http://www.math.boun.edu.tr/instructors/gurel/symbols-a4.pdf

see also:
  _bibtex_undiacritic.py rules
  http://github.com/clld/clld/blob/master/clld/lib/bibtex.py
  http://github.com/clld/clld/blob/master/clld/lib/latex.py
  http://github.com/mcmtroffaes/latexcodec/blob/develop/latexcodec/codec.py#L97

"""
from __future__ import print_function
import re
import unicodedata

from six import text_type, unichr, PY2
import latexcodec

__all__ = [
    'u_escape', 'u_unescape', 'latex_to_utf8', 'ulatex_postprocess', 'ulatex_preprocess']

LATEX_TABLE = {
    u'COMBINING DOT BELOW': br'\textsubdot',
    u'COMBINING MACRON': [br'\=', br'\textbar'],
    u'COMBINING MACRON BELOW': [br'\textsubbar', br'\textsubline'],
    u'COMBINING COMMA BELOW': [br'\,', br'\cb'],
    u'COMBINING TILDE BELOW': br'\textsubtilde',
    u'COMBINING CARON': br'\v',
    u'COMBINING CARON BELOW': br'\textsubwedge',
    u'COMBINING CIRCUMFLEX ACCENT': br'\^',
    u'COMBINING ACUTE ACCENT': br"\'",
    u'COMBINING GRAVE ACCENT': br"\`",
    u'COMBINING DIAERESIS': br'\"',
    u'COMBINING LATIN SMALL LETTER C': br'\textsuperscript{c}',
    u'COMBINING RING BELOW': br'\textsubring',
    u'COMBINING OGONEK': br'\textpolhook',
    u'COMBINING TILDE': br'\~',
    u'COMBINING DOUBLE GRAVE ACCENT': br'\textdoublegrave',
    u'COMBINING BREVE BELOW': br'\textsubu',
    u'COMBINING BRIDGE BELOW': br'\textsubbridge',

    # FIXME: may clash with commands from the textipa package!!
    u'COMBINING HORN': br'\;',
    u'COMBINING DOT ABOVE': [br'\.', br'\textdot', br'\:'],
    # FIXME: that's what textipa provides: but not how it is used in sala.bib!
    #u'LATIN SMALL LETTER N WITH RETROFLEX HOOK': [br'\:{n}', br'\:n'],

    u'MASCULINE ORDINAL INDICATOR': br'\textordmasculine',
    u'LEFT SINGLE QUOTATION MARK': [br'\textquoteleft', br'\grq'],
    u'SINGLE LOW-9 QUOTATION MARK': br'\glq',
    u'RIGHT SINGLE QUOTATION MARK': br'\textquoteright',
    u'NOT SIGN': br'\textlnot',
    u'VULGAR FRACTION THREE QUARTERS': br'\textthreequarters',
    u'LATIN SMALL LETTER A WITH RING ABOVE': br'\r{a}',
    u'LATIN CAPITAL LETTER A WITH RING ABOVE': br'\r{A}',
    u'MODIFIER LETTER SMALL W': br'\textsuperscript{w}',
    u'SUPERSCRIPT ONE': br'\textsuperscript{1}',
    u'SUPERSCRIPT TWO': br'\textsuperscript{2}',
    u'SUPERSCRIPT THREE': br'\textsuperscript{3}',
    u'SUPERSCRIPT FOUR': br'\textsuperscript{4}',
    u'SUPERSCRIPT LATIN SMALL LETTER N': br'\textsuperscript{n}',
    u'PERCENT SIGN': br'\%',
    u'LATIN SMALL LETTER D WITH HOOK': [br'\texthtd', br'\texthooktopd', br'\textrhooktopd', '\!d'],
    u'LATIN SMALL LETTER D WITH TAIL': br'\:{d}',
    u'LATIN SMALL LETTER B WITH HOOK': [br'\texthtb', br'\texthooktopb'],
    u'LATIN SMALL LETTER N WITH LEFT HOOK': br'\textltailn',
    u'LATIN SMALL LETTER S WITH HOOK': br'\textrtails',
    u'LATIN SMALL LETTER ETH': [br'\dh', br'\textdh'],
    u'LATIN CAPITAL LETTER ETH': br'\DH',
    u'LATIN SMALL LETTER ENG': [br'\ng', br'\textipa{N}'],
    u'LATIN CAPITAL LETTER ENG': br'\NG',
    u'LATIN SMALL LETTER OPEN O': br'\textopeno',
    u'LATIN CAPITAL LETTER THORN': br'\TH',
    u'LATIN SMALL LETTER THORN': [br'\textthorn', br'\th'],
    u'LATIN LETTER GLOTTAL STOP': br'\textglotstop',
    u'LATIN LETTER INVERTED GLOTTAL STOP': br'\textrevglotstop',
    u'LATIN SMALL LETTER SCHWA': br'\textschwa',
    u'LATIN SMALL LETTER R WITH TAIL': br'\textrtailr',
    u'LATIN SMALL LETTER BARRED O': br'\texttheta',
    u'LATIN SMALL LETTER O WITH HORN': br'\ohorn',
    u'MODIFIER LETTER GLOTTAL STOP': br'\textraiseglotstop',
    u'LATIN SMALL LETTER I WITH STROKE': br'\textbari',
    u'LATIN SMALL LETTER E WITH TILDE': br'\~e',
    u'INVERTED QUESTION MARK': br'\textquestiondown',
    u'INVERTED EXCLAMATION MARK': [br'\textexclamationdown', br'\textexclamdown'],
    u'HORIZONTAL ELLIPSIS': br'\dots',
    u'RIGHT DOUBLE QUOTATION MARK': br'\textquotedblright',
    u'LEFT DOUBLE QUOTATION MARK': br'\textquotedblleft',
    u'LEFT-POINTING DOUBLE ANGLE QUOTATION MARK': br'\guillemotleft',
    u'RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK': br'\guillemotright',
    u'LATIN LETTER ALVEOLAR CLICK': [br'\textdoublebarpipe', br'\textdoublebarpipevar'],
    u'PLUS-MINUS SIGN': br'\plusminus',
    u'LATIN SMALL LETTER E WITH MACRON AND ACUTE': br'\textacutemacron e',
    u'LATIN SMALL LETTER E WITH OGONEK': br'\textpolhook e',
    u'LATIN SMALL LETTER A WITH OGONEK': br'\c{a}',
    u'LATIN SMALL LETTER TURNED M': br'\textturnm',
    u'MODIFIER LETTER TRIANGULAR COLON': br'\textlengthmark',
    u'LESS-THAN SIGN': br'\textless',
    u'GREATER-THAN SIGN': br'\textgreater',
    # a bit of a misuse, but that's how it is used in goba.bib:
    u'LATIN CROSS': br'\textbarpipe',
    # again a mis-use:
    u'LATIN SMALL LETTER OPEN E': br'\textepsilon',
    u'LATIN SMALL LETTER GAMMA': [br'\textgamma', br'\gamma'],
    u'LATIN SMALL LETTER TURNED V': br'\textturnv',
    u'GREEK SMALL LETTER BETA': br'\textbeta',
    u'GREEK CAPITAL LETTER ETA': br'\textEta',
    u'GREEK SMALL LETTER OMEGA': br'\textomega',
    u'LATIN SMALL LETTER UPSILON': br'\textupsilon',
    u'LATIN SMALL LETTER ESH': br'\textesh',
    u'CYRILLIC SMALL LETTER HARD SIGN': br'\texthardsign',
    u'EURO SIGN': br'\eurosign',
    u'DEGREE SIGN': br'\circ',
    u'VERTICAL LINE': br'\textvertline',
    u'DOUBLE VERTICAL LINE': br'\textdoublevertline',
    u'LATIN SMALL LETTER T WITH RETROFLEX HOOK': br'\:{t}',
    u'LATIN SMALL LETTER H WITH STROKE': br'\textcrh',
    u'LATIN SMALL LETTER REVERSED OPEN E': br'\textrevepsilon',
    u'LATIN SMALL LETTER PHI': br'\textphi',
    u'GREEK SMALL LETTER CHI': br'\textchi',
}

_LC_TABLES = (
    latexcodec.codec._LATEX_UNICODE_TABLE, latexcodec.codec._ULATEX_UNICODE_TABLE)

if PY2:
    assert u'\xe4'.encode('latex') == r'\"a'
    assert r'\"a'.decode('latex') == u'\xe4'

    for unicode_text, latex_text in sorted(
            list(LATEX_TABLE.items()), key=lambda t: t[0], reverse=True):
        if not isinstance(latex_text, (list, tuple)):
            latex_text = [latex_text]
        uchar = unicodedata.lookup(unicode_text)
        for table in _LC_TABLES:
            for lt in latex_text:
                table.register(uchar, lt)


numcharref_patterns = [
    re.compile('\\\\?&#(?P<dec>[0-9]+);'),
    re.compile('\?\[\\\\u\s*(?P<dec>[0-9]+)\]')]

command_patterns = [
    re.compile('\\\\%s \{(?P<arg>[^\}]+)\}' % name)
    for name in ['url', 'emph', 'textit', 'texttt']]


def numcharref_repl(m):
    return unichr(int(m.group('dec')))


def ulatex_preprocess(s):
    # In the post-processing step we rely on all combining characters having been
    # inserted based on latex markup (because we then reorder the string).
    # Thus we have to make sure there are no combining characters in the original string.
    s = unicodedata.normalize('NFC', s.decode('utf8'))
    for c in s:
        if unicodedata.combining(c):  # pragma: no cover
            print(s)
            raise ValueError
    return s


language_tags = {
    'latin': 'lat',
    'zh': 'zho',
    'hindi': 'hin',
    'eng': 'eng',
    'viet': 'vie',
    'tib': 'bod',
    'skt': 'san',  # Sanskrit
    'gujarati': 'guj',
    'pacoh': 'pac',
    'thai': 'tha',
    'dutch': 'nld',
    'burm': 'mya',
    'dan': 'dan',
    'norw': 'nor',
    'oldkhmer': 'qok',
    'ital': 'ita',
    'santali': 'sat',
    'span': 'spa',
    'germ': 'deu',
    'fr': 'fra',
    'rus': 'rus'}
language_tag_pattern = re.compile(
    u"\\\\(?P<tag>%s)\s+" % '|'.join(list(language_tags.keys())))


def recode_language_tags(s):
    def repl(m):
        return '[%s] ' % language_tags[m.group('tag')]
    return language_tag_pattern.sub(repl, s)


def ulatex_postprocess(s, debracket=re.compile(u"\{([^\}]+)\}")):
    if not isinstance(s, text_type):  # pragma: no cover
        s = s.decode('ascii')
    for p in command_patterns:
        s = p.sub(lambda m: m.group('arg'), s)
    s = debracket.sub(u"\\1", debracket.sub(u"\\1", s.replace('{}', '')))
    n = []
    comb = []
    for c in s:
        if unicodedata.combining(c):
            comb.append(c)
        else:
            n.append(c)
            while comb:
                n.append(comb.pop())
    s = u''.join(n)
    for p in numcharref_patterns:
        s = p.sub(numcharref_repl, s)
    for cmd in ['relax', 'it', 'em', 'textsc', 'cite', 'citet']:
        s = s.replace('\\' + cmd + ' ', '')
    s = recode_language_tags(s)
    return unicodedata.normalize('NFC', s)


def ulatex_decode(s):
    return ulatex_postprocess(ulatex_preprocess(s).decode('ulatex'))



# coding=utf8
from __future__ import unicode_literals
import re
import os

from clldutils.path import Path, as_posix, walk
from clldutils.misc import UnicodeMixin, cached_property
from clldutils.inifile import INI
from clldutils.declenum import EnumSymbol
import pycountry
from termcolor import colored
from newick import Node

from pyglottolog import util
from pyglottolog import languoids
from pyglottolog import references
from pyglottolog import objects

ISO_CODE_PATTERN = re.compile('[a-z]{3}$')


class Glottolog(UnicodeMixin):
    countries = [objects.Country(c.alpha_2, c.name) for c in pycountry.countries]

    def __init__(self, repos=None):
        self.repos = Path(repos) if repos else util.DATA_DIR
        self.tree = self.repos.joinpath('languoids', 'tree')

    def __unicode__(self):
        return '<Glottolog repos at %s>' % self.repos

    def build_path(self, *comps):
        build_dir = self.repos.joinpath('build')
        if not build_dir.exists():
            build_dir.mkdir()
        return build_dir.joinpath(*comps)

    def references_path(self, *comps):
        return self.repos.joinpath('references', *comps)

    def languoids_path(self, *comps):
        return self.repos.joinpath('languoids', *comps)

    @cached_property()
    def iso(self):
        return util.get_iso(self.build_path())

    @property
    def glottocodes(self):
        return objects.Glottocodes(self.languoids_path('glottocodes.json'))

    @property
    def ftsindex(self):
        return self.build_path('whoosh')

    def languoid(self, id_):
        if isinstance(id_, languoids.Languoid):
            return id_

        if ISO_CODE_PATTERN.match(id_):
            for d in walk(self.tree, mode='dirs'):
                l = languoids.Languoid.from_dir(d)
                if l.iso_code == id_:
                    return l
        else:
            for d in walk(self.tree, mode='dirs'):
                if d.name == id_:
                    return languoids.Languoid.from_dir(d)

    def languoids(self, ids=None, maxlevel=objects.Level.dialect):
        nodes = {}

        for dirpath, dirnames, filenames in os.walk(as_posix(self.tree)):
            dp = Path(dirpath)
            if dp.name in nodes and nodes[dp.name][2] > maxlevel:
                del dirnames[:]

            for dirname in dirnames:
                if ids is None or dirname in ids:
                    lang = languoids.Languoid.from_dir(dp.joinpath(dirname), nodes=nodes)
                    if lang.level <= maxlevel:
                        yield lang

    def languoids_by_code(self):
        """
        Returns a `dict` mapping the three major language code schemes 
        (Glottocode, ISO code, and Harald's NOCODE_s) to Languoid objects.
        """
        res = {}
        for lang in self.languoids():
            res[lang.id] = lang
            if lang.hid:
                res[lang.hid] = lang
            if lang.iso:
                res[lang.iso] = lang
        return res

    def ascii_tree(self, start, maxlevel=None):
        _ascii_node(self.languoid(start), 0, True, maxlevel, '')

    def newick_tree(self, start):
        return _newick_node(self.languoid(start)).newick

    @cached_property()
    def bibfiles(self):
        return references.BibFiles(
            self, INI.from_file(self.references_path('BIBFILES.ini'), interpolation=None))

    @cached_property()
    def hhtypes(self):
        return references.HHTypes(
            INI.from_file(self.references_path('hhtype.ini'), interpolation=None))

    @cached_property()
    def triggers(self):
        res = {'inlg': [], 'lgcode': []}
        for lang in self.languoids():
            for type_ in res:
                if lang.cfg.has_option('triggers', type_):
                    label = '%s [%s]' % (lang.name, lang.hid or lang.id)
                    res[type_].extend([util.Trigger(type_, label, text)
                                       for text in lang.cfg.getlist('triggers', type_)])
        return res

    @cached_property()
    def macroarea_map(self):
        res = {}
        for lang in self.languoids():
            ma = lang.macroareas[0].value if lang.macroareas else ''
            res[lang.id] = ma
            if lang.iso:
                res[lang.iso] = ma
            if lang.hid:
                res[lang.hid] = ma
        return res


def _newick_node(l):
    n = Node(name="'{0} [{1}]'".format(l.name, l.id))
    for nn in sorted(l.children, key=lambda nn: nn.name):
        n.add_descendant(_newick_node(nn))
    return n


def _ascii_node(n, level, last, maxlevel, prefix):
    if maxlevel:
        if (isinstance(maxlevel, EnumSymbol) and n.level > maxlevel) or \
                (not isinstance(maxlevel, EnumSymbol) and level > maxlevel):
            return
    s = '\u2514' if last else '\u251c'
    s += '\u2500 '

    if not level:
        for i, node in enumerate(n.ancestors):
            util.sprint('{0}{1}{2} [{3}]', prefix, s if i else '', node.name, node.id)
            prefix = '   ' + prefix

    nprefix = prefix + ('   ' if last else '\u2502  ')

    color = 'red' if not level else (
        'green' if n.level == objects.Level.language else (
            'blue' if n.level == objects.Level.dialect else None))

    util.sprint(
        '{0}{1}{2} [{3}]',
        prefix,
        s if level else (s if n.ancestors else ''),
        colored(n.name, color) if color else n.name,
        colored(n.id, color) if color else n.id)
    for i, c in enumerate(sorted(n.children, key=lambda nn: nn.name)):
        _ascii_node(c, level + 1, i == len(n.children) - 1, maxlevel, nprefix)



# coding: utf8
from __future__ import unicode_literals, print_function, division
import re

roman_map = {'m': 1000, 'd': 500, 'c': 100, 'l': 50, 'x': 10, 'v': 5, 'i': 1}
rerom = re.compile("(\d+)")


def introman(i):
    iz = {v: k for k, v in roman_map.items()}
    x = ""
    for v, c in sorted(iz.items(), reverse=True):
        q, r = divmod(i, v)
        if q == 4 and c != 'm':
            x = x + c + iz[5 * v]
        else:
            x += ''.join(c for _ in range(q))
        i = r
    return x


def romanint(r):
    i = 0
    prev = 10000
    for c in r:
        zc = roman_map[c]
        if zc > prev:
            i = i - 2 * prev + zc
        else:
            i += zc
        prev = zc
    return i


def roman(x):
    return rerom.sub(lambda o: introman(int(o.group(1))), x).upper()



# coding: utf8
from __future__ import unicode_literals, print_function, division

from six import text_type, PY2
from mock import Mock, patch
from clldutils.testing import capture
from clldutils.path import copytree

from pyglottolog.tests.util import WithApi


class Tests(WithApi):
    def _args(self, *args):
        self.log = Mock()
        return Mock(repos=self.api, args=list(args), log=self.log)

    def test_show(self):
        from pyglottolog.commands import show

        with capture(show, self._args('**a:key**')) as out:
            self.assertIn('@misc'.encode('utf8') if PY2 else '@misc', out)

        with capture(show, self._args('a:key')) as out:
            self.assertIn('@misc'.encode('utf8') if PY2 else '@misc', out)

        with capture(show, self._args('abcd1236')) as out:
            self.assertIn('Classification'.encode('utf8') if PY2 else 'Classificat', out)

    def test_edit(self):
        from pyglottolog.commands import edit

        with patch('pyglottolog.commands.subprocess', Mock()):
            edit(self._args('abcd1236'))

    def test_create(self):
        from pyglottolog.commands import create

        with capture(create, self._args('abcd1234', 'new name', 'language')) as out:
            self.assertIn('Info written', out)
            self.assertIn(
                'new name', [c.name for c in self.api.languoid('abcd1234').children])

    def test_fts(self):
        from pyglottolog.commands import refindex, refsearch, langindex, langsearch

        refindex(self._args())
        with capture(refsearch, self._args('Harzani year:1334')) as out:
            self.assertIn("'Abd-al-'Ali Karang", out)

        langindex(self._args())
        with capture(langsearch, self._args('id:abcd*')) as out:
            self.assertIn("abcd1234", out)

        with capture(langsearch, self._args('classification')) as out:
            self.assertIn("abcd1234", out)

    def test_metadata(self):
        from pyglottolog.commands import metadata

        with capture(metadata, self._args()) as out:
            self.assertIn("longitude", out)

    def test_lff(self):
        from pyglottolog.commands import tree2lff, lff2tree

        tree2lff(self._args())
        with self.api.build_path('dff.txt').open(encoding='utf8') as fp:
            dfftxt = fp.read().replace('dialect', 'Dialect Name')
        with self.api.build_path('dff.txt').open('w', encoding='utf8') as fp:
            fp.write(dfftxt)
        with capture(lff2tree, self._args()) as out:
            self.assertIn('git status', out)
        self.assertEqual(self.api.languoid('abcd1236').name, 'Dialect Name')
        tree2lff(self._args())
        with self.api.build_path('dff.txt').open(encoding='utf8') as fp:
            self.assertEqual(dfftxt, fp.read())

    def test_index(self):
        from pyglottolog.commands import index

        index(self._args())
        self.assertEqual(
            len(list(self.repos.joinpath('languoids').glob('*.md'))), 7)

    def test_tree(self):
        from pyglottolog.commands import tree, ParserError

        with self.assertRaises(ParserError):
            tree(self._args())

        with self.assertRaises(ParserError):
            tree(self._args('xyz'))

        with capture(tree, self._args('abc', 'language')) as out:
            if not isinstance(out, text_type):
                out = out.decode('utf8')
            self.assertIn('language', out)
            self.assertNotIn('dialect', out)

    def test_newick(self):
        from pyglottolog.commands import newick

        with capture(newick, self._args('abcd1235')) as out:
            self.assertIn('language', out)

    def test_check(self):
        from pyglottolog.commands import check

        with capture(check, self._args('refs')) as _:
            pass

        with capture(check, self._args()) as out:
            self.assertIn('family', out)
        for call in self.log.error.call_args_list:
            self.assertIn('unregistered glottocode', call[0][0])
        self.assertEqual(self.log.error.call_count, 4)

        copytree(
            self.api.tree.joinpath('abcd1234', 'abcd1235'),
            self.api.tree.joinpath('abcd1235'))

        with capture(check, self._args()) as _:
            self.assertIn(
                'duplicate glottocode',
                ''.join(c[0][0] for c in self.log.error.call_args_list))
            self.assertEqual(self.log.error.call_count, 6)

    def test_monster(self):
        from pyglottolog.commands import bib

        if not PY2:  # pragma: no cover
            return

        with capture(bib, self._args()) as out:
            assert out



# coding: utf8
"""
Main command line interface of the pyglottolog package.

Like programs such as git, this cli splits its functionality into sub-commands
(see e.g. https://docs.python.org/2/library/argparse.html#sub-commands).
The rationale behind this is that while a lot of different tasks may be triggered using
this cli, most of them require common configuration.

The basic invocation looks like

    glottolog [OPTIONS] <command> [args]

"""
from __future__ import unicode_literals, print_function
import sys

from clldutils.clilib import ArgumentParserWithLogging, ParserError
from clldutils.path import copytree, rmtree, remove

from pyglottolog.languoids import Glottocode
from pyglottolog.api import Glottolog
from pyglottolog import commands
assert commands


def recode(args):
    """Assign a new glottocode to an existing languoid.

    glottolog recode <code>
    """
    lang = args.repos.languoid(args.args[0])
    if not lang:
        raise ParserError('languoid not found')
    lang.id = Glottocode.from_name(lang.name)
    new_dir = lang.dir.parent.joinpath(lang.id)
    copytree(lang.dir, new_dir)
    lang.write_info(new_dir)
    remove(new_dir.joinpath('%s.ini' % args.args[0]))
    rmtree(lang.dir)
    print("%s -> %s" % (args.args[0], lang.id))


def main():  # pragma: no cover
    parser = ArgumentParserWithLogging('pyglottolog')
    parser.add_argument(
        '--repos',
        help="path to glottolog data repository",
        type=Glottolog,
        default=Glottolog())
    sys.exit(parser.main())



# coding: utf8
from __future__ import unicode_literals
import os
from collections import defaultdict

from clldutils.misc import UnicodeMixin
from clldutils.path import Path
from clldutils.inifile import INI

from pyglottolog.objects import (
    Glottocode, EthnologueComment, Reference, Level, Country, Macroarea,
    EndangermentStatus, ClassificationComment, ISORetirement,
)

INFO_FILENAME = 'md.ini'


class Languoid(UnicodeMixin):
    """
    Info on languoids is encoded in the ini files and in the directory hierarchy.
    This class provides access to all of it.
    """
    section_core = 'core'

    def __init__(self, cfg, lineage=None, id_=None, directory=None, tree=None):
        """

        :param cfg:
        :param lineage: list of ancestors, given as (id, name) pairs.
        """
        assert (id_ and tree) or directory
        if id_ is None:
            id_ = Glottocode(directory.name)
        lineage = lineage or []
        assert all([
            Glottocode.pattern.match(id)
            and Level.get(level) for name, id, level in lineage])
        self.lineage = [(name, id, Level.get(level)) for name, id, level in lineage]
        self.cfg = cfg
        self.dir = directory or tree.joinpath(*[id for name, id, _ in self.lineage])
        self._id = id_

    @classmethod
    def from_dir(cls, directory, nodes=None, **kw):
        if nodes is None:
            nodes = {}
        cfg = INI.from_file(directory.joinpath(INFO_FILENAME), interpolation=None)

        lineage = []
        for parent in directory.parents:
            id_ = parent.name
            assert id_ != directory.name
            if not Glottocode.pattern.match(id_):
                # we ignore leading non-languoid-dir path components.
                break

            if id_ not in nodes:
                l = Languoid.from_dir(parent, nodes=nodes)
                nodes[id_] = (l.name, l.id, l.level)
            lineage.append(nodes[id_])

        res = cls(cfg, list(reversed(lineage)), directory=directory)
        nodes[res.id] = (res.name, res.id, res.level)
        return res

    @classmethod
    def from_name_id_level(cls, tree, name, id, level, **kw):
        cfg = INI(interpolation=None)
        cfg.read_dict(dict(core=dict(name=name)))
        res = cls(cfg, kw.pop('lineage', []), id_=Glottocode(id), tree=tree)
        res.level = Level.get(level)
        for k, v in kw.items():
            setattr(res, k, v)
        return res

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return '<%s %s>' % (self.level.name.capitalize(), self.id)

    def __unicode__(self):
        return '%s [%s]' % (self.name, self.id)

    def _set(self, key, value):
        if value is None and key in self.cfg[self.section_core]:
            del self.cfg[self.section_core][key]
        else:
            self.cfg.set(self.section_core, key, value)

    def _get(self, key, type_=None):
        res = self.cfg.get(self.section_core, key, fallback=None)
        if type_ and res:
            return type_(res)
        return res

    def write_info(self, outdir=None):
        outdir = outdir or self.dir
        if not isinstance(outdir, Path):
            outdir = Path(outdir)
        if outdir.name != self.id:
            outdir = outdir.joinpath(self.id)
        if not outdir.exists():
            outdir.mkdir()
        fname = outdir.joinpath(INFO_FILENAME)
        self.cfg.write(fname)
        if os.linesep == '\n':
            with fname.open(encoding='utf8') as fp:
                text = fp.read()
            with fname.open('w', encoding='utf8') as fp:
                fp.write(text.replace('\n', '\r\n'))
        return fname

    # -------------------------------------------------------------------------
    # Accessing info of a languoid
    # -------------------------------------------------------------------------
    @property
    def glottocode(self):
        return self._id

    @property
    def id(self):
        return self._id

    @property
    def category(self):
        fid = self.lineage[0][1] if self.lineage else None
        category_map = defaultdict(
            lambda: 'Spoken L1 Language',
            **{
                'book1242': 'Bookkeeping',
                'unat1236': 'Unattested',
                'uncl1493': 'Unclassifiable',
                'sign1238': 'Sign Language',
                'arti1236': 'Artificial Language',
                'spee1234': 'Speech Register',
                'pidg1258': 'Pidgin',
                'mixe1287': 'Mixed Language',
            })
        if self.level == Level.language:
            return category_map[fid]
        cat = self.level.name.capitalize()
        if self.level == Level.family:
            if self.id.startswith('unun9') or \
                    self.id in category_map or fid in category_map:
                cat = 'Pseudo ' + cat
        return cat

    @property
    def isolate(self):
        return self.level == Level.language and not self.lineage

    def children_from_nodemap(self, nodes):
        # A faster alternative to `children` when the relevant languoids have already been
        # read from disc.
        return [nodes[d.name] for d in self.dir.iterdir() if d.is_dir()]

    @property
    def children(self):
        return [Languoid.from_dir(d) for d in self.dir.iterdir() if d.is_dir()]

    def ancestors_from_nodemap(self, nodes):
        # A faster alternative to `ancestors` when the relevant languoids have already
        # been read from disc.
        return [nodes[l[1]] for l in self.lineage]

    @property
    def ancestors(self):
        res = []
        for parent in self.dir.parents:
            id_ = parent.name
            if Glottocode.pattern.match(id_):
                res.append(Languoid.from_dir(parent))
            else:
                # we ignore leading non-languoid-dir path components.
                break
        return list(reversed(res))

    @property
    def parent(self):
        ancestors = self.ancestors
        return ancestors[-1] if ancestors else None

    @property
    def family(self):
        ancestors = self.ancestors
        return ancestors[0] if ancestors else None

    @property
    def names(self):
        if 'altnames' in self.cfg:
            return {k: self.cfg.getlist('altnames', k) for k in self.cfg['altnames']}
        return {}

    @property
    def identifier(self):
        if 'identifier' in self.cfg:
            return self.cfg['identifier']
        return {}

    @property
    def sources(self):
        if self.cfg.has_option('sources', 'glottolog'):
            return Reference.from_list(self.cfg.getlist('sources', 'glottolog'))
        return []

    @sources.setter
    def sources(self, refs):
        assert all(isinstance(r, Reference) for r in refs)
        self.cfg.set('sources', 'glottolog', ['{0}'.format(ref) for ref in refs])

    @property
    def endangerment(self):
        if 'status' in self.cfg[self.section_core]:
            res = self.cfg.get(self.section_core, 'status')
            if res:
                return EndangermentStatus.get(res.lower())

    @endangerment.setter
    def endangerment(self, value):
        value = value.lower() if hasattr(value, 'lower') else value
        self._set('status', EndangermentStatus.get(value).name)

    @property
    def classification_comment(self):
        if 'classification' in self.cfg:
            cfg = self.cfg['classification']
            return ClassificationComment(
                family=cfg.get('family'),
                familyrefs=self.cfg.getlist('classification', 'familyrefs'),
                sub=cfg.get('sub'),
                subrefs=self.cfg.getlist('classification', 'subrefs'))

    @property
    def ethnologue_comment(self):
        section = 'hh_ethnologue_comment'
        if section in self.cfg:
            return EthnologueComment(**self.cfg[section])

    @property
    def macroareas(self):
        return [Macroarea.get(n)
                for n in self.cfg.getlist(self.section_core, 'macroareas')]

    @macroareas.setter
    def macroareas(self, value):
        assert isinstance(value, (list, tuple)) \
            and all(o in list(Macroarea) for o in value)
        self._set('macroareas', ['{0}'.format(ma) for ma in value])

    @property
    def countries(self):
        return [Country.from_text(n)
                for n in self.cfg.getlist(self.section_core, 'countries')]

    @countries.setter
    def countries(self, value):
        assert isinstance(value, (list, tuple)) \
               and all(isinstance(o, Country) for o in value)
        self._set('countries', ['{0}'.format(c) for c in value])

    @property
    def name(self):
        return self._get('name')

    @name.setter
    def name(self, value):
        self._set('name', value)

    @property
    def latitude(self):
        return self._get('latitude', float)

    @latitude.setter
    def latitude(self, value):
        self._set('latitude', float(value))

    @property
    def longitude(self):
        return self._get('longitude', float)

    @longitude.setter
    def longitude(self, value):
        self._set('longitude', float(value))

    @property
    def hid(self):
        return self._get('hid')

    @hid.setter
    def hid(self, value):
        self._set('hid', value)

    @property
    def level(self):
        return self._get('level', Level.get)

    @level.setter
    def level(self, value):
        self._set('level', Level.get(value).name)

    @property
    def iso(self):
        return self._get('iso639-3')

    @iso.setter
    def iso(self, value):
        self._set('iso639-3', value)

    @property
    def iso_code(self):
        return self._get('iso639-3')

    @iso_code.setter
    def iso_code(self, value):
        self._set('iso639-3', value)

    @property
    def iso_retirement(self):
        if 'iso_retirement' in self.cfg:
            kw = self.cfg['iso_retirement']
            if 'comment' in kw:
                kw['comment'] = self.cfg.gettext('iso_retirement', 'comment')
            return ISORetirement(**self.cfg['iso_retirement'])

    @property
    def fname(self):
        return self.dir.joinpath(INFO_FILENAME)



# coding: utf8
"""
Create a BibTeX file listing records for each past ISO 639-3 change request.

http://www-01.sil.org/iso639-3/chg_requests.asp?order=CR_Number&chg_status=past
"""
from __future__ import unicode_literals, print_function, division
from itertools import groupby

import requests
from bs4 import BeautifulSoup as bs

from clldutils.source import Source


BASE_URL = "http://www-01.sil.org/iso639-3"


def change_request_as_source(id_, rows, ref_ids):
    title = "Change Request Number {0}: ".format(id_)
    title += ", ".join(
        "{0} {1} [{2}]".format(
            r['Outcome/Effective date'].split('20')[0].strip().lower(),
            r['Change Type'].lower(),
            r['Affected Identifier'])
        for r in rows)
    date = None
    for row in rows:
        parts = row['Outcome/Effective date'].split('20')
        if len(parts) > 1:
            if date:
                assert date == parts[1].strip()
            else:
                date = parts[1].strip()
    if date:
        title += ' ({0})'.format(date)
    fields = {
        'number': id_,
        'title': title,
        'howpublished': BASE_URL + "/chg_detail.asp?id=" + id_,
        'address': "Dallas",
        'author': "ISO 639-3 Registration Authority",
        'publisher': "SIL International",
        'url': BASE_URL + "/cr_files/{0}.pdf".format(id_),
        'year': id_.split('-')[0],
        'hhtype': "overview",
        'lgcode': ', '.join(
            "{0} [{1}]".format(r['Language Name'].strip(), r['Affected Identifier'])
            for r in rows),
        'src': "iso6393",
    }
    if id_ in ref_ids and ref_ids[id_]:
        fields['glottolog_ref_id'] = ref_ids[id_]
    return Source('misc', id_, **fields)


def iter_change_requests(log):
    def parse_row(tr, coltag):
        return [td.get_text() for td in tr.find_all(coltag)]

    url = BASE_URL + "/chg_requests.asp"
    log.info('downloading {0} ...'.format(url))
    res = requests.get(url, params=dict(order='CR_Number', chg_status='past'))
    log.info('HTTP {0}'.format(res.status_code))
    table = bs(res.content, 'html.parser').find('table')
    cols = None
    for i, tr in enumerate(table.find_all('tr')):
        if i == 0:
            cols = parse_row(tr, 'th')
        else:
            yield dict(zip(cols, parse_row(tr, 'td')))


def bibtex(api, log):
    bib = api.bibfiles['iso6393.bib']
    glottolog_ref_ids = bib.glottolog_ref_id_map
    i = -1

    with bib.fname.open('w', encoding='utf8') as fp:
        for i, (id_, rows) in enumerate(groupby(
                iter_change_requests(log), lambda c: c['CR Number'])):
            fp.write(
                change_request_as_source(id_, list(rows), glottolog_ref_ids).bibtex())
            fp.write('\n\n')
    log.info('bibtex written to {0}'.format(bib.fname))
    return i + 1



# coding: utf8
from __future__ import unicode_literals, print_function, division

from clldutils.path import read_text, write_text

from pyglottolog.tests.util import WithApi


class Tests(WithApi):
    def test_Entry(self):
        from pyglottolog.references import Entry

        self.assertEqual(Entry.lgcodes(None), [])

    def test_HHTypes(self):
        hht = self.api.hhtypes
        self.assertEqual(hht['grammar'].rank, 17)

        prev = None
        for t in hht:
            if prev:
                self.assertGreater(prev, t)
            prev = t

        self.assertEqual(len(hht), 2)
        self.assertIn('rank', repr(hht[0]))
        self.assertEqual(hht.parse('grammar'), ['grammar'])

    def test_BibFile(self):
        bibfile = self.api.bibfiles['a.bib']
        self.assertEqual(bibfile['a:key'].type, 'misc')
        self.assertEqual(bibfile['s:Andalusi:Turk'].key, 's:Andalusi:Turk')

        for entry in bibfile.iterentries():
            if entry.key == 'key':
                self.assertEqual(len(list(entry.iterlanguoids({'abc': 1}))), 1)

        with self.assertRaises(KeyError):
            _ = bibfile['xyz']

        self.assertEqual(len(list(bibfile.iterentries())), 3)

        lines = [line for line in read_text(bibfile.fname).split('\n')
                 if not line.strip().startswith('glottolog_ref_id')]
        write_text(self.tmp_path('a.bib'), '\n'.join(lines))
        bibfile.update(self.tmp_path('a.bib'))
        self.assertEqual(len(list(bibfile.iterentries())), 3)

        bibfile.update(self.api.bibfiles['b.bib'].fname)
        self.assertEqual(len(list(bibfile.iterentries())), 1)

        def visitor(entry):
            entry.fields['new_field'] = 'a'

        bibfile.visit(visitor=visitor)
        for entry in bibfile.iterentries():
            self.assertIn('new_field', entry.fields)

        bibfile.visit(visitor=lambda e: True)
        self.assertEqual(len(bibfile.keys()), 0)



# coding: utf8
from __future__ import unicode_literals, print_function, division

from clldutils.path import Path, copytree
from clldutils.testing import WithTempDir

from pyglottolog.api import Glottolog


class WithApi(WithTempDir):
    def setUp(self):
        WithTempDir.setUp(self)
        self.repos = self.tmp_path('repos')
        copytree(Path(__file__).parent.joinpath('data'), self.repos)
        self.api = Glottolog(self.repos)



# coding: utf8
from __future__ import unicode_literals, print_function, division

from six import PY2
from clldutils.testing import capture
from pyglottolog.tests.util import WithApi


class Tests(WithApi):
    def test_main(self):
        from pyglottolog.monster import compile

        if not PY2:  # pragma: no cover
            return

        with capture(compile, self.api) as out:
            self.assertEqual(len(out.splitlines()), 44)
            self.assertIn('2 splitted', out)
            self.assertIn('2 merged', out)



# _bibfiles_db.py - load bibfiles into sqlite3, hash, assign ids (split/merge)

import sqlite3
import difflib
import operator
import itertools
import contextlib
import collections
import logging

from six import string_types, viewkeys
from clldutils.dsv import UnicodeWriter
from clldutils import jsonlib
from clldutils.path import remove

from pyglottolog.util import unique, group_first
from pyglottolog.monsterlib import _bibtex

log = logging.getLogger('pyglottolog')
UNION_FIELDS = {'fn', 'asjp_name', 'isbn'}
IGNORE_FIELDS = {'crossref', 'numnote', 'glotto_id'}


class Database(object):
    """Bibfile collection parsed into an sqlite3 file."""

    @classmethod
    def from_bibfiles(cls, bibfiles, filename, rebuild=False):
        """If needed, (re)build the db from the bibfiles, hash, split/merge."""
        if filename.exists():
            if not rebuild:
                self = cls(filename, bibfiles)
                if self.is_uptodate():
                    return self
            remove(filename)

        self = cls(filename, bibfiles)
        with self.connect(async=True) as conn:
            create_tables(conn)
            with conn:
                import_bibfiles(conn, bibfiles)
            entrystats(conn)
            fieldstats(conn)

            with conn:
                generate_hashes(conn)
            hashstats(conn)
            hashidstats(conn)

            with conn:
                assign_ids(conn)

        return self

    def __init__(self, filename, bibfiles):
        self.filename = filename.as_posix()
        self._bibfiles = bibfiles

    def is_uptodate(self, bibfiles=None, verbose=False):
        """Does the db have the same filenames, sizes, and mtimes as bibfiles?"""
        with self.connect() as conn:
            return compare_bibfiles(conn, bibfiles or self._bibfiles, verbose=verbose)

    def recompute(self, hashes=True, reload_priorities=True, verbose=True):
        """Call _libmonster.keyid for all entries, splits/merges -> new ids."""
        with self.connect(async=True) as conn:
            if hashes:
                with conn:
                    generate_hashes(conn)
                hashstats(conn)
                hashidstats(conn)
            if reload_priorities:
                source = None if reload_priorities is True else reload_priorities
                with conn:
                    update_priorities(conn, self._bibfiles)
            with conn:
                assign_ids(conn, verbose=verbose)

    def to_bibfile(self, filename, encoding='utf-8', ):
        _bibtex.save(self.merged(), filename.as_posix(), sortkey=None, encoding=encoding)

    def to_csvfile(self, filename):
        """Write a CSV file with one row for each entry in each bibfile."""
        with self.connect() as conn:
            cursor = conn.execute(
                'SELECT filename, bibkey, hash, cast(id AS text) AS id '
                'FROM entry ORDER BY lower(filename), lower(bibkey), hash, id')
            with UnicodeWriter(filename) as writer:
                writer.writerow([col[0] for col in cursor.description])
                for row in cursor:
                    writer.writerow(row)

    def to_replacements(self, filename):
        """Write a JSON file with 301s from merged glottolog_ref_ids."""
        with self.connect() as conn:
            conn.row_factory = sqlite3.Row
            cursor = conn.execute('SELECT refid AS id, id AS replacement '
                'FROM entry WHERE id != refid ORDER BY id')
            pairs = map(dict, cursor)
        with jsonlib.update(filename, default=[], indent=4) as repls:
            repls.extend(pairs)

    def to_hhmapping(self):
        with self.connect() as conn:
            assert allid(conn)
            query = 'SELECT bibkey, id FROM entry WHERE filename = ?'
            return dict(conn.execute(query, ('hh.bib',)))

    def trickle(self):
        """Write new/changed glottolog_ref_ids back into the bibfiles."""
        bibfiles = self._bibfiles
        if not self.is_uptodate(verbose=True):
            raise RuntimeError('trickle with an outdated db')  # pragma: no cover
        with self.connect() as conn:
            filenames = conn.execute('SELECT name FROM file WHERE EXISTS '
                '(SELECT 1 FROM entry WHERE filename = name '
                'AND id != coalesce(refid, -1)) ORDER BY name').fetchall()
            for f, in filenames:
                b = bibfiles[f]
                entries = b.load()
                cursor = conn.execute('SELECT bibkey, cast(refid AS text), cast(id AS text) '
                    'FROM entry WHERE filename = ? AND id != coalesce(refid, -1) '
                    'ORDER BY lower(bibkey)', (f,))
                added = changed = 0
                for bibkey, refid, new in cursor:
                    entrytype, fields = entries[bibkey]
                    old = fields.pop('glottolog_ref_id', None)
                    assert old == refid
                    if old is None:
                        added += 1
                    else:
                        changed += 1
                    fields['glottolog_ref_id'] = new
                print('%d changed %d added in %s' % (changed, added, b.id))
                b.save(entries)

    def merged(self):
        """Yield merged (bibkey, (entrytype, fields)) entries."""
        for (id, hash), grp in self:
            entrytype, fields = self._merged_entry(grp)
            fields['glottolog_ref_id'] = id
            yield hash, (entrytype, fields)

    def connect(self, close=True, async=False):
        conn = sqlite3.connect(self.filename)
        if async:
            conn.execute('PRAGMA synchronous = OFF')
            conn.execute('PRAGMA journal_mode = MEMORY')
        if close:
            conn = contextlib.closing(conn)
        return conn

    def __iter__(self, chunksize=100):
        with self.connect() as conn:
            assert allid(conn)

            allpriority, = conn.execute('SELECT NOT EXISTS '
                '(SELECT 1 FROM entry WHERE NOT EXISTS (SELECT 1 FROM file '
                'WHERE name = filename))').fetchone()
            assert allpriority

            assert onetoone(conn)

            get_id_hash, get_field = operator.itemgetter(0, 1), operator.itemgetter(2)
            for first, last in windowed(conn, 'id', chunksize):
                cursor = conn.execute('SELECT e.id, e.hash, v.field, v.value, v.filename, v.bibkey '
                    'FROM entry AS e '
                    'JOIN file AS f ON e.filename = f.name '
                    'JOIN value AS v ON e.filename = v.filename AND e.bibkey = v.bibkey '
                    'LEFT JOIN field AS d ON v.filename = d.filename AND v.field = d.field '
                    'WHERE e.id BETWEEN ? AND ? '
                    'ORDER BY e.id, v.field, coalesce(d.priority, f.priority) DESC, v.filename, v.bibkey',
                    (first, last))
                for id_hash, grp in itertools.groupby(cursor, get_id_hash):
                    yield (id_hash, [(field, [(vl, fn, bk) for id, hs, fd, vl, fn, bk in g])
                        for field, g in itertools.groupby(grp, get_field)])

    def __getitem__(self, key):
        """Entry by (fn, bk) or merged entry by refid (old grouping) or hash (current grouping)."""
        if not isinstance(key, (tuple, int)) and not isinstance(key, string_types):
            raise ValueError  # pragma: no cover
        with self.connect() as conn:
            if isinstance(key, tuple):
                filename, bibkey = key
                entrytype, fields = self._entry(conn, filename, bibkey)
            else:
                grp = self._entrygrp(conn, key)
                entrytype, fields = self._merged_entry(grp)
            return key, (entrytype, fields)

    @staticmethod
    def _entry(conn, filename, bibkey, raw=False):
        cursor = conn.execute('SELECT field, value FROM value '
            'WHERE filename = ? AND bibkey = ? ', (filename, bibkey))
        fields = dict(cursor)
        if not fields:
            raise KeyError((filename, bibkey))
        if raw:
            return fields
        entrytype = fields.pop('ENTRYTYPE')
        return entrytype, fields

    @staticmethod
    def _merged_entry(grp, union=UNION_FIELDS, ignore=IGNORE_FIELDS, raw=False):
        # TODO: consider implementing (a subset of?) onlyifnot logic:
        # {'address': 'publisher', 'lgfamily': 'lgcode', 'publisher': 'school', 'journal': 'booktitle'}
        fields = {field: values[0][0] if field not in union
            else ', '.join(unique(vl for vl, fn, bk in values))
            for field, values in grp if field not in ignore}
        fields['src'] = ', '.join(sorted(set(fn.partition('.bib')[0]
            for field, values in grp for vl, fn, bk in values)))
        fields['srctrickle'] = ', '.join(sorted(set('%s#%s' % (fn.partition('.bib')[0], bk)
            for field, values in grp for vl, fn, bk in values)))
        if raw:
            return fields
        entrytype = fields.pop('ENTRYTYPE')
        return entrytype, fields

    @staticmethod
    def _entrygrp(conn, key, get_field=operator.itemgetter(0)):
        col = 'refid' if isinstance(key, int) else 'hash'
        cursor = conn.execute(('SELECT v.field, v.value, v.filename, v.bibkey '
            'FROM entry AS e '
            'JOIN file AS f ON e.filename = f.name '
            'JOIN value AS v ON e.filename = v.filename AND e.bibkey = v.bibkey '
            'LEFT JOIN field AS d ON v.filename = d.filename AND v.field = d.field '
            'WHERE %s = ? '
            'ORDER BY v.field, coalesce(d.priority, f.priority) DESC, v.filename, v.bibkey'
            ) % col, (key,))
        grp = [(field, [(vl, fn, bk) for fd, vl, fn, bk in g])
            for field, g in itertools.groupby(cursor, get_field)]
        if not grp:
            raise KeyError(key)
        return grp

    def stats(self, field_files=False):
        with self.connect() as conn:
            entrystats(conn)
            fieldstats(conn, field_files)
            hashstats(conn)
            hashidstats(conn)

    @staticmethod
    def print_group(conn, group):
        for row in group:
            print(row)
        for row in group:
            print('\t%r, %r, %r, %r' % hashfields(conn, row[-2], row[-1]))

    def show_splits(self):
        with self.connect() as conn:
            cursor = conn.execute('SELECT refid, hash, filename, bibkey '
            'FROM entry AS e WHERE EXISTS (SELECT 1 FROM entry '
            'WHERE refid = e.refid AND hash != e.hash) '
            'ORDER BY refid, hash, filename, bibkey')
            for refid, group in group_first(cursor):
                self.print_group(conn, group)
                old = self._merged_entry(self._entrygrp(conn, refid), raw=True)
                cand = [(hs, self._merged_entry(self._entrygrp(conn, hs), raw=True))
                    for hs in unique(hs for ri, hs, fn, bk in group)]
                new = min(cand, key=lambda p: distance(old, p[1]))[0]
                print('-> %s\n' % new)

    def show_merges(self):
        with self.connect() as conn:
            cursor = conn.execute('SELECT hash, refid, filename, bibkey '
            'FROM entry AS e WHERE EXISTS (SELECT 1 FROM entry '
            'WHERE hash = e.hash AND refid != e.refid) '
            'ORDER BY hash, refid DESC, filename, bibkey')
            for hash, group in group_first(cursor):
                self.print_group(conn, group)
                new = self._merged_entry(self._entrygrp(conn, hash), raw=True)
                cand = [(ri, self._merged_entry(self._entrygrp(conn, ri), raw=True))
                    for ri in unique(ri for hs, ri, fn, bk in group)]
                old = min(cand, key=lambda p: distance(new, p[1]))[0]
                print('-> %s\n' % old)

    def _show(self, sql):
        with self.connect() as conn:
            cursor = conn.execute(sql)
            for hash, group in group_first(cursor):
                self.print_group(conn, group)
                print()

    def show_identified(self):
        self._show(
            'SELECT hash, refid, filename, bibkey '
            'FROM entry AS e WHERE EXISTS (SELECT 1 FROM entry '
            'WHERE refid IS NULL AND hash = e.hash) '
            'AND EXISTS (SELECT 1 FROM entry '
            'WHERE refid IS NOT NULL AND hash = e.hash) '
            'ORDER BY hash, refid IS NOT NULL, refid, filename, bibkey')

    def show_combined(self):
        self._show(
            'SELECT hash, filename, bibkey '
            'FROM entry AS e WHERE refid IS NULL AND EXISTS (SELECT 1 FROM entry '
            'WHERE refid IS NULL AND hash = e.hash '
            'AND (filename != e.filename OR bibkey != e.bibkey)) '
            'ORDER BY hash, filename, bibkey')


def create_tables(conn, page_size=32768):
    if page_size is not None:
        conn.execute('PRAGMA page_size = %d' % page_size)
    conn.execute('CREATE TABLE file ('
        'name TEXT NOT NULL, '
        'size INTEGER NOT NULL, '
        'mtime DATETIME NOT NULL, '
        'priority INTEGER NOT NULL, '
        'PRIMARY KEY (name))')
    conn.execute('CREATE TABLE field ('
        'filename TEXT NOT NULL, '
        'field TEXT NOT NULL, '
        'priority INTEGER NOT NULL, '
        'PRIMARY KEY (filename, field), '
        'FOREIGN KEY (filename) REFERENCES file(name))')
    conn.execute('CREATE TABLE entry ('
        'filename TEXT NOT NULL, '
        'bibkey TEXT NOT NULL, '
        'refid INTEGER, '  # old glottolog_ref_id from the bibfiles (previous hash groupings)
        'hash TEXT, '      # current groupings, m:n with refid (splits/merges)
        'srefid INTEGER, ' # split-resolved refid (every srefid maps to exactly one hash)
        'id INTEGER, '     # new glottolog_ref_id to save into the bibfiles (current hash groupings)
        'PRIMARY KEY (filename, bibkey), '
        'FOREIGN KEY (filename) REFERENCES file(name))')
    conn.execute('CREATE INDEX ix_refid ON entry(refid)')
    conn.execute('CREATE INDEX ix_hash ON entry(hash)')
    conn.execute('CREATE INDEX ix_srefid ON entry(srefid)')
    conn.execute('CREATE INDEX ix_id ON entry(id)')
    conn.execute('CREATE TABLE value ('
        'filename TEXT NOT NULL, '
        'bibkey TEXT NOT NULL, '
        'field TEXT NOT NULL, '
        'value TEXT NOT NULL, '
        'PRIMARY KEY (filename, bibkey, field), '
        'FOREIGN KEY (filename, bibkey) REFERENCES entry(filename, bibkey))')
 

def import_bibfiles(conn, bibfiles):
    log.info('importing bibfiles into a new db')
    for b in bibfiles:
        conn.execute('INSERT INTO file (name, size, mtime, priority)'
            'VALUES (?, ?, ?, ?)', (b.fname.name, b.size, b.mtime, b.priority))
        for e in b.iterentries():
            bibkey, entrytype, fields = e.key, e.type, e.fields
            conn.execute(
                'INSERT INTO entry (filename, bibkey, refid) VALUES (?, ?, ?)',
                (b.fname.name, bibkey, fields.get('glottolog_ref_id')))
            fields = itertools.chain([('ENTRYTYPE', entrytype)], fields.items())
            conn.executemany(
                'INSERT INTO value (filename, bibkey, field, value) VALUES (?, ?, ?, ?)',
                ((b.fname.name, bibkey, field, value) for field, value in fields))


def update_priorities(conn, bibfiles):
    inini = {b.fname.name for b in bibfiles}
    indb = {filename for filename, in conn.execute('SELECT name FROM file')}
    assert inini == indb
    for b in bibfiles:
        conn.execute(
            'UPDATE file SET priority = ? WHERE NAME = ?',
            (b.priority, b.fname.name))
    print('\n'.join('%d\t%s' % pn for pn in conn.execute(
        'SELECT priority, name FROM file ORDER BY priority DESC, name')))


def compare_bibfiles(conn, bibfiles, verbose=False):
    ondisk = collections.OrderedDict((b.fname.name, (b.size, str(b.mtime)))
        for b in bibfiles)
    indb = collections.OrderedDict((name, (size, mtime))
        for name, size, mtime in
        conn.execute('SELECT name, size, mtime FROM file ORDER BY name'))
    if dict(ondisk) == dict(indb):
        return True
    if verbose:
        print('missing in db: %s' % [o for o in ondisk if o not in indb])
        print('missing on disk: %s' % [i for i in indb if i not in ondisk])
        print('differing in size/mtime: %s' % [o for o in ondisk
            if o in indb and ondisk[o] != indb[o]])
    return False


def allid(conn):
    result, = conn.execute('SELECT NOT EXISTS (SELECT 1 FROM entry '
        'WHERE id IS NULL)').fetchone()
    return result


def onetoone(conn):
    result, = conn.execute('SELECT NOT EXISTS '
        '(SELECT 1 FROM entry AS e WHERE EXISTS (SELECT 1 FROM entry '
        'WHERE hash = e.hash AND id != e.id '
        'OR id = e.id AND hash != e.hash))').fetchone()
    return result


def entrystats(conn):
    log.info('entry stats:\n' + '\n'.join('%s %d' % (f, n) for f, n in conn.execute(
        'SELECT filename, count(*) FROM entry GROUP BY filename')))
    log.info('%d entries total' % conn.execute('SELECT count(*) FROM entry').fetchone())


def fieldstats(conn, with_files=False):
    if with_files:
        print('\n'.join('%d\t%s\t%s' % (n, f, b) for f, n, b in conn.execute(
            'SELECT field, count(*) AS n, replace(group_concat(DISTINCT filename), ",", ", ") '
            'FROM value GROUP BY field ORDER BY n DESC, field')))
    else:
        print('\n'.join('%d\t%s' % (n, f) for f, n in conn.execute(
            'SELECT field, count(*) AS n '
            'FROM value GROUP BY field ORDER BY n DESC, field')))


def windowed_entries(conn, chunksize):
    for filename, in conn.execute('SELECT name FROM file ORDER BY name'):
        cursor = conn.execute('SELECT bibkey FROM entry WHERE filename = ? '
            'ORDER BY bibkey', (filename,))
        while True:
            bibkeys = cursor.fetchmany(chunksize)
            if not bibkeys:
                cursor.close()
                break
            (first,), (last,) = bibkeys[0], bibkeys[-1]
            yield filename, first, last


def hashfields(conn, filename, bibkey):
    # also: extra_hash, volume (if not journal, booktitle, or series)
    cursor = conn.execute('SELECT field, value FROM value '
        "WHERE field IN ('author', 'editor', 'year', 'title') "
        'AND filename = ? AND bibkey = ? ', (filename, bibkey))
    fields = dict(cursor)
    return tuple(fields.get(f) for f in ('author', 'editor', 'year', 'title'))


def generate_hashes(conn):
    from _libmonster import wrds, keyid

    words = collections.Counter()
    cursor = conn.execute('SELECT value FROM value WHERE field = ?', ('title',))
    while True:
        rows = cursor.fetchmany(10000)
        if not rows:
            break
        for title, in rows:
            words.update(wrds(title))
    # TODO: consider dropping stop words/hapaxes from freq. distribution
    print('%d title words (from %d tokens)' % (len(words), sum(words.values())))

    get_bibkey = operator.itemgetter(0)
    for filename, first, last in windowed_entries(conn, 500):
        rows = conn.execute('SELECT bibkey, field, value FROM value '
            'WHERE filename = ? AND bibkey BETWEEN ? AND ? '
            'AND field != ? ORDER BY bibkey', (filename, first, last, 'ENTRYTYPE'))
        conn.executemany('UPDATE entry SET hash = ? WHERE filename = ? AND bibkey = ?',
            ((keyid({k: v for b, k, v in grp}, words), filename, bibkey)
            for bibkey, grp in itertools.groupby(rows, get_bibkey)))


def hashstats(conn):
    print('%d\tdistinct keyids (from %d total)' % conn.execute(
        'SELECT count(DISTINCT hash), count(hash) FROM entry').fetchone())
    print('\n'.join('%d\t%s (from %d distinct of %d total)' % row
        for row in conn.execute('SELECT coalesce(c2.unq, 0), '
        'c1.filename, c1.dst, c1.tot FROM (SELECT filename, '
        'count(hash) AS tot, count(DISTINCT hash) AS dst  '
        'FROM entry GROUP BY filename) AS c1 LEFT JOIN '
        '(SELECT filename, count(DISTINCT hash) AS unq '
        'FROM entry AS e WHERE NOT EXISTS (SELECT 1 FROM entry '
        'WHERE hash = e.hash AND filename != e.filename) '
        'GROUP BY filename) AS c2 ON c1.filename = c2.filename '
        'ORDER BY c1.filename')))
    print('%d\tin multiple files' % conn.execute('SELECT count(*) FROM '
        '(SELECT 1 FROM entry GROUP BY hash '
        'HAVING COUNT(DISTINCT filename) > 1)').fetchone())


def hashidstats(conn):
    print('\n'.join('1 keyid %d glottolog_ref_ids: %d' % (hash_nid, n)
        for (hash_nid, n) in conn.execute(
        'SELECT hash_nid, count(*) AS n FROM '
        '(SELECT count(DISTINCT refid) AS hash_nid FROM entry WHERE hash IS NOT NULL '
        'GROUP BY hash HAVING count(DISTINCT refid) > 1) '
        'GROUP BY hash_nid ORDER BY n desc')))
    print('\n'.join('1 glottolog_ref_id %d keyids: %d' % (id_nhash, n)
        for (id_nhash, n) in conn.execute(
        'SELECT id_nhash, count(*) AS n FROM '
        '(SELECT count(DISTINCT hash) AS id_nhash FROM entry WHERE refid IS NOT NULL '
        'GROUP BY refid HAVING count(DISTINCT hash) > 1) '
        'GROUP BY id_nhash ORDER BY n desc')))


def windowed(conn, col, chunksize):
    query = 'SELECT DISTINCT %(col)s FROM entry ORDER BY %(col)s' % {'col': col}
    cursor = conn.execute(query)
    while True:
        rows = cursor.fetchmany(chunksize)
        if not rows:
            cursor.close()
            break
        (first,), (last,) = rows[0], rows[-1]
        yield first, last


def assign_ids(conn, verbose=False):
    merged_entry, entrygrp = Database._merged_entry, Database._entrygrp

    allhash, = conn.execute('SELECT NOT EXISTS (SELECT 1 FROM entry '
        'WHERE hash IS NULL)').fetchone()
    assert allhash

    print('%d entries' % conn.execute('UPDATE entry SET id = NULL, srefid = refid').rowcount)

    # resolve splits: srefid = refid only for entries from the most similar hash group
    nsplit = 0
    cursor = conn.execute('SELECT refid, hash, filename, bibkey FROM entry AS e '
        'WHERE EXISTS (SELECT 1 FROM entry WHERE refid = e.refid AND hash != e.hash) '
        'ORDER BY refid, hash, filename, bibkey')
    for refid, group in group_first(cursor):
        old = merged_entry(entrygrp(conn, refid), raw=True)
        nsplit += len(group)
        cand = [(hs, merged_entry(entrygrp(conn, hs), raw=True))
            for hs in unique(hs for ri, hs, fn, bk in group)]
        new = min(cand, key=lambda p: distance(old, p[1]))[0]
        separated = conn.execute('UPDATE entry SET srefid = NULL WHERE refid = ? AND hash != ?',
            (refid, new)).rowcount
        if verbose:
            for row in group:
                print(row)
            for ri, hs, fn, bk in group:
                print('\t%r, %r, %r, %r' % hashfields(conn, fn, bk))
            print('-> %s' % new)
            print('%d: %d separated from %s\n' % (refid, separated, new))
    print('%d splitted' % nsplit)
    
    nosplits, = conn.execute('SELECT NOT EXISTS (SELECT 1 FROM entry AS e '
        'WHERE EXISTS (SELECT 1 FROM entry WHERE srefid = e.srefid AND hash != e.hash))').fetchone()
    assert nosplits

    # resolve merges: id = srefid of the most similar srefid group
    nmerge = 0
    cursor = conn.execute('SELECT hash, srefid, filename, bibkey FROM entry AS e '
        'WHERE EXISTS (SELECT 1 FROM entry WHERE hash = e.hash AND srefid != e.srefid) '
        'ORDER BY hash, srefid DESC, filename, bibkey')
    for hash, group in group_first(cursor):
        new = merged_entry(entrygrp(conn, hash), raw=True)
        nmerge += len(group)
        cand = [(ri, merged_entry(entrygrp(conn, ri), raw=True))
            for ri in unique(ri for hs, ri, fn, bk in group)]
        old = min(cand, key=lambda p: distance(new, p[1]))[0]
        merged = conn.execute('UPDATE entry SET id = ? WHERE hash = ? AND srefid != ?',
            (old, hash, old)).rowcount
        if verbose:
            for row in group:
                print(row)
            for hs, ri, fn, bk in group:
                print('\t%r, %r, %r, %r' % hashfields(conn, fn, bk))
            print('-> %s' % old)
            print('%s: %d merged into %d\n' % (hash, merged, old))
    print('%d merged' % nmerge)

    # unchanged entries
    print('%d unchanged' % conn.execute('UPDATE entry SET id = srefid '
        'WHERE id IS NULL AND srefid IS NOT NULL').rowcount)

    nomerges, = conn.execute('SELECT NOT EXISTS (SELECT 1 FROM entry AS e '
        'WHERE EXISTS (SELECT 1 FROM entry WHERE hash = e.hash AND id != e.id))').fetchone()
    assert nomerges

    # identified
    print('%d identified (new/separated)' % conn.execute('UPDATE entry '
        'SET id = (SELECT id FROM entry AS e WHERE e.hash = entry.hash AND e.id IS NOT NULL) '
        'WHERE refid IS NULL AND id IS NULL AND EXISTS '
        '(SELECT 1 FROM entry AS e WHERE e.hash = entry.hash AND e.id IS NOT NULL)').rowcount)

    # assign new ids to hash groups of separated/new entries
    nextid, = conn.execute('SELECT coalesce(max(refid), 0) + 1 FROM entry').fetchone()
    cursor = conn.execute('SELECT hash FROM entry WHERE id IS NULL GROUP BY hash ORDER BY hash')
    print('%d new ids (new/separated)' % conn.executemany('UPDATE entry SET id = ? WHERE hash = ?',
        ((id, hash) for id, (hash,) in enumerate(cursor, nextid))).rowcount)

    assert allid(conn)
    assert onetoone(conn)

    # supersede relation
    superseded, = conn.execute('SELECT count(*) FROM entry WHERE id != srefid').fetchone()
    print('%d supersede pairs' % superseded)


def distance(left, right, weight={'author': 3, 'year': 3, 'title': 3, 'ENTRYTYPE': 2}):
    """Simple measure of the difference between two bibtex-field dicts."""
    if not (left or right):
        return 0.0

    keys = viewkeys(left) & viewkeys(right)
    if not keys:
        return 1.0

    weights = {k: weight.get(k, 1) for k in keys}
    ratios = (
        w * difflib.SequenceMatcher(None, left[k], right[k]).ratio()
        for k, w in weights.items())
    return 1 - (sum(ratios) / sum(weights.values()))


def _test_merge():  # pragma: no cover
    import sqlalchemy as sa

    engine = sa.create_engine('postgresql://postgres@/overrides')
    metadata = sa.MetaData()
    overrides = sa.Table(
        'overrides', metadata,
        sa.Column('hash', sa.Text, primary_key=True),
        sa.Column('field', sa.Text, primary_key=True),
        sa.Column('file1', sa.Text, primary_key=True),
        sa.Column('bibkey1', sa.Text, primary_key=True),
        sa.Column('file2', sa.Text, primary_key=True),
        sa.Column('bibkey2', sa.Text, primary_key=True),
        sa.Column('value1', sa.Text),
        sa.Column('value2', sa.Text))
    metadata.drop_all(engine)
    metadata.create_all(engine)
    insert_ov = overrides.insert(bind=engine).execute

    for hash, grp in Database():
        for field, values in grp:
            if field in UNION_FIELDS:
                continue
            value1, file1, bibkey1 = values[0]
            for value2, file2, bibkey2 in values[1:]:
                if value1.lower() != value2.lower():
                    insert_ov(hash=hash, field=field, value1=value1, value2=value2,
                        file1=file1, bibkey1=bibkey1, file2=file2, bibkey2=bibkey2)

    query = sa.select([
        overrides.c.file1, overrides.c.file2, sa.func.count().label('n')
        ]).where(overrides.c.file1 != overrides.c.file2)\
        .group_by(overrides.c.file1, overrides.c.file2)\
        .order_by(sa.literal_column('n'), overrides.c.file1, overrides.c.file2)

    print('\n'.join('%d\t%s\t%s' % (n, f1, f2)
        for f1, f2, n in engine.execute(query)))
